# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

""" Test Z2Symmetries """

from test.python.opflow import QiskitOpflowTestCase

from qiskit.opflow import PauliSumOp, TaperedPauliSumOp, Z2Symmetries
from qiskit.quantum_info import Pauli, SparsePauliOp


class TestZ2Symmetries(QiskitOpflowTestCase):
    """Z2Symmetries tests."""

    def test_find_Z2_symmetries(self):
        """test for find_Z2_symmetries"""

        qubit_op = PauliSumOp.from_list(
            [
                ("II", -1.0537076071291125),
                ("IZ", 0.393983679438514),
                ("ZI", -0.39398367943851387),
                ("ZZ", -0.01123658523318205),
                ("XX", 0.1812888082114961),
            ]
        )
        z2_symmetries = Z2Symmetries.find_Z2_symmetries(qubit_op)
        self.assertEqual(z2_symmetries.symmetries, [Pauli("ZZ")])
        self.assertEqual(z2_symmetries.sq_paulis, [Pauli("IX")])
        self.assertEqual(z2_symmetries.sq_list, [0])
        self.assertEqual(z2_symmetries.tapering_values, None)

        tapered_op = z2_symmetries.taper(qubit_op)[1]
        self.assertEqual(tapered_op.z2_symmetries.symmetries, [Pauli("ZZ")])
        self.assertEqual(tapered_op.z2_symmetries.sq_paulis, [Pauli("IX")])
        self.assertEqual(tapered_op.z2_symmetries.sq_list, [0])
        self.assertEqual(tapered_op.z2_symmetries.tapering_values, [-1])

        z2_symmetries.tapering_values = [-1]
        primitive = SparsePauliOp.from_list(
            [
                ("I", -1.0424710218959303),
                ("Z", -0.7879673588770277),
                ("X", -0.18128880821149604),
            ]
        )
        expected_op = TaperedPauliSumOp(primitive, z2_symmetries)
        self.assertEqual(tapered_op, expected_op)

    def test_taper_empty_operator(self):
        """Test tapering of empty operator"""
        z2_symmetries = Z2Symmetries(
            symmetries=[Pauli("IIZI"), Pauli("IZIZ"), Pauli("ZIII")],
            sq_paulis=[Pauli("IIXI"), Pauli("IIIX"), Pauli("XIII")],
            sq_list=[1, 0, 3],
            tapering_values=[1, -1, -1],
        )
        empty_op = PauliSumOp.from_list([("IIII", 0.0)])
        tapered_op = z2_symmetries.taper(empty_op)
        expected_op = PauliSumOp.from_list([("I", 0.0)])
        self.assertEqual(tapered_op, expected_op)

    def test_truncate_tapered_op(self):
        """Test setting cutoff tolerances for the tapered operator works."""
        qubit_op = PauliSumOp.from_list(
            [
                ("II", -1.0537076071291125),
                ("IZ", 0.393983679438514),
                ("ZI", -0.39398367943851387),
                ("ZZ", -0.01123658523318205),
                ("XX", 0.1812888082114961),
            ]
        )
        z2_symmetries = Z2Symmetries.find_Z2_symmetries(qubit_op)
        z2_symmetries.tol = 0.2  # removes the X part of the tapered op which is < 0.2

        tapered_op = z2_symmetries.taper(qubit_op)[1]
        primitive = SparsePauliOp.from_list(
            [
                ("I", -1.0424710218959303),
                ("Z", -0.7879673588770277),
            ]
        )
        expected_op = TaperedPauliSumOp(primitive, z2_symmetries)
        self.assertEqual(tapered_op, expected_op)

    def test_twostep_tapering(self):
        """Test the two-step tapering"""
        qubit_op = PauliSumOp.from_list(
            [
                ("II", -1.0537076071291125),
                ("IZ", 0.393983679438514),
                ("ZI", -0.39398367943851387),
                ("ZZ", -0.01123658523318205),
                ("XX", 0.1812888082114961),
            ]
        )
        z2_symmetries = Z2Symmetries.find_Z2_symmetries(qubit_op)
        tapered_op = z2_symmetries.taper(qubit_op)

        tapered_op_firststep = z2_symmetries.convert_clifford(qubit_op)
        tapered_op_secondstep = z2_symmetries.taper_clifford(tapered_op_firststep)
        self.assertEqual(tapered_op, tapered_op_secondstep)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""A wrapper class for the purposes of validating modifications to
QuantumCircuit.data while maintaining the interface of a python list."""

from collections.abc import MutableSequence
from typing import Tuple, Iterable, Optional

from .exceptions import CircuitError
from .instruction import Instruction
from .quantumregister import Qubit
from .classicalregister import Clbit


class CircuitInstruction:
    """A single instruction in a :class:`.QuantumCircuit`, comprised of the :attr:`operation` and
    various operands.

    .. warning::

        This is a lightweight internal class and there is minimal error checking; you must respect
        the type hints when using it.  It is the user's responsibility to ensure that direct
        mutations of the object do not invalidate the types, nor the restrictions placed on it by
        its context.  Typically this will mean, for example, that :attr:`qubits` must be a sequence
        of distinct items, with no duplicates.
    """

    __slots__ = ("operation", "qubits", "clbits", "_legacy_format_cache")

    operation: Instruction
    """The logical operation that this instruction represents an execution of."""
    qubits: Tuple[Qubit, ...]
    """A sequence of the qubits that the operation is applied to."""
    clbits: Tuple[Clbit, ...]
    """A sequence of the classical bits that this operation reads from or writes to."""

    def __init__(
        self,
        operation: Instruction,
        qubits: Iterable[Qubit] = (),
        clbits: Iterable[Clbit] = (),
    ):
        self.operation = operation
        self.qubits = tuple(qubits)
        self.clbits = tuple(clbits)
        self._legacy_format_cache = None

    def copy(self) -> "CircuitInstruction":
        """Return a shallow copy of the :class:`CircuitInstruction`."""
        return self.__class__(
            operation=self.operation,
            qubits=self.qubits,
            clbits=self.clbits,
        )

    def replace(
        self,
        operation: Optional[Instruction] = None,
        qubits: Optional[Iterable[Qubit]] = None,
        clbits: Optional[Iterable[Clbit]] = None,
    ) -> "CircuitInstruction":
        """Return a new :class:`CircuitInstruction` with the given fields replaced."""
        return self.__class__(
            operation=self.operation if operation is None else operation,
            qubits=self.qubits if qubits is None else qubits,
            clbits=self.clbits if clbits is None else clbits,
        )

    def __repr__(self):
        return (
            f"{type(self).__name__}("
            f"operation={self.operation!r}"
            f", qubits={self.qubits!r}"
            f", clbits={self.clbits!r}"
            ")"
        )

    def __eq__(self, other):
        if isinstance(other, type(self)):
            # Ordered from fastest comparisons to slowest.
            return (
                self.clbits == other.clbits
                and self.qubits == other.qubits
                and self.operation == other.operation
            )
        if isinstance(other, tuple):
            return self._legacy_format == other
        return NotImplemented

    # Legacy tuple-like interface support.
    #
    # For a best attempt at API compatibility during the transition to using this new class, we need
    # the interface to behave exactly like the old 3-tuple `(inst, qargs, cargs)` if it's treated
    # like that via unpacking or similar.  That means that the `parameters` field is completely
    # absent, and the qubits and clbits must be converted to lists.

    @property
    def _legacy_format(self):
        if self._legacy_format_cache is None:
            # The qubits and clbits were generally stored as lists in the old format, and various
            # places assume that they will certainly be lists.
            self._legacy_format_cache = (self.operation, list(self.qubits), list(self.clbits))
        return self._legacy_format_cache

    def __getitem__(self, key):
        return self._legacy_format[key]

    def __iter__(self):
        return iter(self._legacy_format)

    def __len__(self):
        return 3


class QuantumCircuitData(MutableSequence):
    """A wrapper class for the purposes of validating modifications to
    QuantumCircuit.data while maintaining the interface of a python list."""

    def __init__(self, circuit):
        self._circuit = circuit

    def __getitem__(self, i):
        return self._circuit._data[i]

    def __setitem__(self, key, value):
        # For now (Terra 0.21), the `QuantumCircuit.data` setter is meant to perform validation, so
        # we do the same qubit checks that `QuantumCircuit.append` would do.
        if isinstance(value, CircuitInstruction):
            operation, qargs, cargs = value.operation, value.qubits, value.clbits
        else:
            # Handle the legacy 3-tuple format.
            operation, qargs, cargs = value
        value = self._resolve_legacy_value(operation, qargs, cargs)
        self._circuit._data[key] = value
        self._circuit._update_parameter_table(value)

    def _resolve_legacy_value(self, operation, qargs, cargs) -> CircuitInstruction:
        """Resolve the old-style 3-tuple into the new :class:`CircuitInstruction` type."""
        if not isinstance(operation, Instruction) and hasattr(operation, "to_instruction"):
            operation = operation.to_instruction()
        if not isinstance(operation, Instruction):
            raise CircuitError("object is not an Instruction.")

        expanded_qargs = [self._circuit.qbit_argument_conversion(qarg) for qarg in qargs or []]
        expanded_cargs = [self._circuit.cbit_argument_conversion(carg) for carg in cargs or []]

        if isinstance(operation, Instruction):
            broadcast_args = list(operation.broadcast_arguments(expanded_qargs, expanded_cargs))
        else:
            broadcast_args = list(
                Instruction.broadcast_arguments(operation, expanded_qargs, expanded_cargs)
            )

        if len(broadcast_args) > 1:
            raise CircuitError(
                "QuantumCircuit.data modification does not support argument broadcasting."
            )

        qargs, cargs = broadcast_args[0]

        self._circuit._check_dups(qargs)
        return CircuitInstruction(operation, tuple(qargs), tuple(cargs))

    def insert(self, index, value):
        self._circuit._data.insert(index, None)
        try:
            self[index] = value
        except CircuitError:
            del self._circuit._data[index]
            raise

    def __iter__(self):
        return iter(self._circuit._data)

    def __delitem__(self, i):
        del self._circuit._data[i]

    def __len__(self):
        return len(self._circuit._data)

    def __cast(self, other):
        return other._circuit._data if isinstance(other, QuantumCircuitData) else other

    def __repr__(self):
        return repr(self._circuit._data)

    def __lt__(self, other):
        return self._circuit._data < self.__cast(other)

    def __le__(self, other):
        return self._circuit._data <= self.__cast(other)

    def __eq__(self, other):
        return self._circuit._data == self.__cast(other)

    def __gt__(self, other):
        return self._circuit._data > self.__cast(other)

    def __ge__(self, other):
        return self._circuit._data >= self.__cast(other)

    def __add__(self, other):
        return self._circuit._data + self.__cast(other)

    def __radd__(self, other):
        return self.__cast(other) + self._circuit._data

    def __mul__(self, n):
        return self._circuit._data * n

    def __rmul__(self, n):
        return n * self._circuit._data

    def sort(self, *args, **kwargs):
        """In-place stable sort. Accepts arguments of list.sort."""
        self._circuit._data.sort(*args, **kwargs)

    def copy(self):
        """Returns a shallow copy of instruction list."""
        return self._circuit._data.copy()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

r"""

.. estimator-desc:

=====================
Overview of Estimator
=====================

Estimator class estimates expectation values of quantum circuits and observables.

An estimator is initialized with an empty parameter set. The estimator is used to
create a :class:`~qiskit.providers.JobV1`, via the
:meth:`qiskit.primitives.Estimator.run()` method. This method is called
with the following parameters

* quantum circuits (:math:`\psi_i(\theta)`): list of (parameterized) quantum circuits
  (a list of :class:`~qiskit.circuit.QuantumCircuit` objects).

* observables (:math:`H_j`): a list of :class:`~qiskit.quantum_info.SparsePauliOp`
  objects.

* parameter values (:math:`\theta_k`): list of sets of values
  to be bound to the parameters of the quantum circuits
  (list of list of float).

The method returns a :class:`~qiskit.providers.JobV1` object, calling
:meth:`qiskit.providers.JobV1.result()` yields the
a list of expectation values plus optional metadata like confidence intervals for
the estimation.

.. math::

    \langle\psi_i(\theta_k)|H_j|\psi_i(\theta_k)\rangle

Here is an example of how the estimator is used.

.. code-block:: python

    from qiskit.primitives import Estimator
    from qiskit.circuit.library import RealAmplitudes
    from qiskit.quantum_info import SparsePauliOp

    psi1 = RealAmplitudes(num_qubits=2, reps=2)
    psi2 = RealAmplitudes(num_qubits=2, reps=3)

    H1 = SparsePauliOp.from_list([("II", 1), ("IZ", 2), ("XI", 3)])
    H2 = SparsePauliOp.from_list([("IZ", 1)])
    H3 = SparsePauliOp.from_list([("ZI", 1), ("ZZ", 1)])

    theta1 = [0, 1, 1, 2, 3, 5]
    theta2 = [0, 1, 1, 2, 3, 5, 8, 13]
    theta3 = [1, 2, 3, 4, 5, 6]

    estimator = Estimator()

    # calculate [ <psi1(theta1)|H1|psi1(theta1)> ]
    job = estimator.run([psi1], [H1], [theta1])
    job_result = job.result() # It will block until the job finishes.
    print(f"The primitive-job finished with result {job_result}"))

    # calculate [ <psi1(theta1)|H1|psi1(theta1)>,
    #             <psi2(theta2)|H2|psi2(theta2)>,
    #             <psi1(theta3)|H3|psi1(theta3)> ]
    job2 = estimator.run([psi1, psi2, psi1], [H1, H2, H3], [theta1, theta2, theta3])
    job_result = job2.result()
    print(f"The primitive-job finished with result {job_result}")
"""

from __future__ import annotations

from abc import abstractmethod
from collections.abc import Iterable, Sequence
from copy import copy
from typing import cast
from warnings import warn

import numpy as np

from qiskit.circuit import Parameter, QuantumCircuit
from qiskit.circuit.parametertable import ParameterView
from qiskit.opflow import PauliSumOp
from qiskit.providers import JobV1 as Job
from qiskit.quantum_info.operators import SparsePauliOp
from qiskit.quantum_info.operators.base_operator import BaseOperator
from qiskit.utils.deprecation import deprecate_arguments, deprecate_function

from .base_primitive import BasePrimitive
from .estimator_result import EstimatorResult
from ..utils import _circuit_key, _observable_key, init_observable


class BaseEstimator(BasePrimitive):
    """Estimator base class.

    Base class for Estimator that estimates expectation values of quantum circuits and observables.
    """

    __hash__ = None

    def __init__(
        self,
        circuits: Iterable[QuantumCircuit] | QuantumCircuit | None = None,
        observables: Iterable[SparsePauliOp] | SparsePauliOp | None = None,
        parameters: Iterable[Iterable[Parameter]] | None = None,
        options: dict | None = None,
    ):
        """
        Creating an instance of an Estimator, or using one in a ``with`` context opens a session that
        holds resources until the instance is ``close()`` ed or the context is exited.

        Args:
            circuits: Quantum circuits that represent quantum states.
            observables: Observables.
            parameters: Parameters of quantum circuits, specifying the order in which values
                will be bound. Defaults to ``[circ.parameters for circ in circuits]``
                The indexing is such that ``parameters[i, j]`` is the j-th formal parameter of
                ``circuits[i]``.
            options: Default options.

        Raises:
            ValueError: For mismatch of circuits and parameters list.
        """
        if circuits is not None or observables is not None or parameters is not None:
            warn(
                "The BaseEstimator `circuits`, `observables`, `parameters` kwarg are deprecated "
                "as of Qiskit Terra 0.22.0 and will be removed no earlier than 3 months after "
                "the release date. You can use the 'run' method to append objects.",
                DeprecationWarning,
                2,
            )
        if isinstance(circuits, QuantumCircuit):
            circuits = (circuits,)
        self._circuits = [] if circuits is None else list(circuits)

        if isinstance(observables, SparsePauliOp):
            observables = (observables,)
        self._observables = [] if observables is None else list(observables)

        # To guarantee that they exist as instance variable.
        # With only dynamic set, the python will not know if the attribute exists or not.
        self._circuit_ids: dict[tuple, int] = self._circuit_ids
        self._observable_ids: dict[tuple, int] = self._observable_ids

        if parameters is None:
            self._parameters = [circ.parameters for circ in self._circuits]
        else:
            self._parameters = [ParameterView(par) for par in parameters]
            if len(self._parameters) != len(self._circuits):
                raise ValueError(
                    f"Different number of parameters ({len(self._parameters)}) and "
                    f"circuits ({len(self._circuits)})"
                )
            for i, (circ, params) in enumerate(zip(self._circuits, self._parameters)):
                if circ.num_parameters != len(params):
                    raise ValueError(
                        f"Different numbers of parameters of {i}-th circuit: "
                        f"expected {circ.num_parameters}, actual {len(params)}."
                    )
        super().__init__(options)

    ################################################################################
    ## METHODS
    ################################################################################
    def run(
        self,
        circuits: Sequence[QuantumCircuit] | QuantumCircuit,
        observables: Sequence[BaseOperator | PauliSumOp] | BaseOperator | PauliSumOp,
        parameter_values: Sequence[Sequence[float]] | Sequence[float] | float | None = None,
        **run_options,
    ) -> Job:
        """Run the job of the estimation of expectation value(s).

        ``circuits``, ``observables``, and ``parameter_values`` should have the same
        length. The i-th element of the result is the expectation of observable

        .. code-block:: python

            obs = observables[i]

        for the state prepared by

        .. code-block:: python

            circ = circuits[i]

        with bound parameters

        .. code-block:: python

            values = parameter_values[i].

        Args:
            circuits: one or more circuit objects.
            observables: one or more observable objects.
            parameter_values: concrete parameters to be bound.
            run_options: runtime options used for circuit execution.

        Returns:
            The job object of EstimatorResult.

        Raises:
            TypeError: Invalid argument type given.
            ValueError: Invalid argument values given.
        """
        # Singular validation
        circuits = self._validate_circuits(circuits)
        observables = self._validate_observables(observables)
        parameter_values = self._validate_parameter_values(
            parameter_values,
            default=[()] * len(circuits),
        )

        # Cross-validation
        self._cross_validate_circuits_parameter_values(circuits, parameter_values)
        self._cross_validate_circuits_observables(circuits, observables)

        # Options
        run_opts = copy(self.options)
        run_opts.update_options(**run_options)

        return self._run(
            circuits,
            observables,
            parameter_values,
            **run_opts.__dict__,
        )

    # This will be comment out after 0.22. (This is necessary for the compatibility.)
    # @abstractmethod
    def _run(
        self,
        circuits: tuple[QuantumCircuit, ...],
        observables: tuple[BaseOperator | PauliSumOp, ...],
        parameter_values: tuple[tuple[float, ...], ...],
        **run_options,
    ) -> Job:
        raise NotImplementedError(
            "_run method is not implemented. This method will be @abstractmethod after 0.22."
        )

    ################################################################################
    ## VALIDATION
    ################################################################################
    @staticmethod
    def _validate_observables(
        observables: Sequence[BaseOperator | PauliSumOp] | BaseOperator | PauliSumOp,
    ) -> tuple[BaseOperator | PauliSumOp, ...]:
        if isinstance(observables, (BaseOperator, PauliSumOp)):
            observables = (observables,)
        elif not isinstance(observables, Sequence) or not all(
            isinstance(obs, (BaseOperator, PauliSumOp)) for obs in observables
        ):
            raise TypeError("Invalid observables, expected Sequence[BaseOperator|PauliSumOp].")
        elif not isinstance(observables, tuple):
            observables = tuple(observables)
        if len(observables) == 0:
            raise ValueError("No observables were provided.")
        return observables

    @staticmethod
    def _cross_validate_circuits_observables(
        circuits: tuple[QuantumCircuit, ...], observables: tuple[BaseOperator | PauliSumOp, ...]
    ) -> None:
        if len(circuits) != len(observables):
            raise ValueError(
                f"The number of circuits ({len(circuits)}) does not match "
                f"the number of observables ({len(observables)})."
            )
        for i, (circuit, observable) in enumerate(zip(circuits, observables)):
            if circuit.num_qubits != observable.num_qubits:
                raise ValueError(
                    f"The number of qubits of the {i}-th circuit ({circuit.num_qubits}) does "
                    f"not match the number of qubits of the {i}-th observable "
                    f"({observable.num_qubits})."
                )

    ################################################################################
    ## DEPRECATED
    ################################################################################
    def __new__(
        cls,
        circuits: Iterable[QuantumCircuit] | QuantumCircuit | None = None,
        observables: Iterable[SparsePauliOp] | SparsePauliOp | None = None,
        parameters: Iterable[Iterable[Parameter]] | None = None,  # pylint: disable=unused-argument
        **kwargs,  # pylint: disable=unused-argument
    ):
        self = super().__new__(cls)
        if circuits is None:
            self._circuit_ids = {}
        elif isinstance(circuits, Iterable):
            circuits = copy(circuits)
            self._circuit_ids = {_circuit_key(circuit): i for i, circuit in enumerate(circuits)}
        else:
            self._circuit_ids = {_circuit_key(circuits): 0}
        if observables is None:
            self._observable_ids = {}
        elif isinstance(observables, Iterable):
            observables = copy(observables)
            self._observable_ids = {
                _observable_key(init_observable(observable)): i
                for i, observable in enumerate(observables)
            }
        else:
            self._observable_ids = {_observable_key(init_observable(observables)): 0}
        return self

    @property
    def circuits(self) -> tuple[QuantumCircuit, ...]:
        """Quantum circuits that represents quantum states.

        Returns:
            The quantum circuits.
        """
        return tuple(self._circuits)

    @property
    def observables(self) -> tuple[SparsePauliOp, ...]:
        """Observables to be estimated.

        Returns:
            The observables.
        """
        return tuple(self._observables)

    @property
    def parameters(self) -> tuple[ParameterView, ...]:
        """Parameters of the quantum circuits.

        Returns:
            Parameters, where ``parameters[i][j]`` is the j-th parameter of the i-th circuit.
        """
        return tuple(self._parameters)

    @deprecate_function(
        "The BaseEstimator.__enter__ method is deprecated as of Qiskit Terra 0.22.0 "
        "and will be removed no sooner than 3 months after the releasedate. "
        "BaseEstimator should be initialized directly.",
    )
    def __enter__(self):
        return self

    @deprecate_function(
        "The BaseEstimator.__call__ method is deprecated as of Qiskit Terra 0.22.0 "
        "and will be removed no sooner than 3 months after the releasedate. "
        "BaseEstimator should be initialized directly.",
    )
    def __exit__(self, *exc_info):
        self.close()

    def close(self):
        """Close the session and free resources"""
        ...

    @deprecate_function(
        "The BaseEstimator.__call__ method is deprecated as of Qiskit Terra 0.22.0 "
        "and will be removed no sooner than 3 months after the releasedate. "
        "Use the 'run' method instead.",
    )
    @deprecate_arguments({"circuit_indices": "circuits", "observable_indices": "observables"})
    def __call__(
        self,
        circuits: Sequence[int | QuantumCircuit],
        observables: Sequence[int | SparsePauliOp],
        parameter_values: Sequence[Sequence[float]] | None = None,
        **run_options,
    ) -> EstimatorResult:
        """Run the estimation of expectation value(s).

        ``circuits``, ``observables``, and ``parameter_values`` should have the same
        length. The i-th element of the result is the expectation of observable

        .. code-block:: python

            obs = self.observables[observables[i]]

        for the state prepared by

        .. code-block:: python

            circ = self.circuits[circuits[i]]

        with bound parameters

        .. code-block:: python

            values = parameter_values[i].

        Args:
            circuits: the list of circuit indices or circuit objects.
            observables: the list of observable indices or observable objects.
            parameter_values: concrete parameters to be bound.
            run_options: Default runtime options used for circuit execution.

        Returns:
            EstimatorResult: The result of the estimator.

        Raises:
            ValueError: For mismatch of object id.
            ValueError: For mismatch of length of Sequence.
        """

        # Support ndarray
        if isinstance(parameter_values, np.ndarray):
            parameter_values = parameter_values.tolist()

        # Allow objects
        circuits = [
            self._circuit_ids.get(_circuit_key(circuit))
            if not isinstance(circuit, (int, np.integer))
            else circuit
            for circuit in circuits
        ]
        if any(circuit is None for circuit in circuits):
            raise ValueError(
                "The circuits passed when calling estimator is not one of the circuits used to "
                "initialize the session."
            )
        observables = [
            self._observable_ids.get(_observable_key(observable))
            if not isinstance(observable, (int, np.integer))
            else observable
            for observable in observables
        ]
        if any(observable is None for observable in observables):
            raise ValueError(
                "The observables passed when calling estimator is not one of the observables used to "
                "initialize the session."
            )

        circuits = cast("list[int]", circuits)
        observables = cast("list[int]", observables)

        # Allow optional
        if parameter_values is None:
            for i in circuits:
                if len(self._circuits[i].parameters) != 0:
                    raise ValueError(
                        f"The {i}-th circuit is parameterised,"
                        "but parameter values are not given."
                    )
            parameter_values = [[]] * len(circuits)

        # Validation
        if len(circuits) != len(observables):
            raise ValueError(
                f"The number of circuits ({len(circuits)}) does not match "
                f"the number of observables ({len(observables)})."
            )
        if len(circuits) != len(parameter_values):
            raise ValueError(
                f"The number of circuits ({len(circuits)}) does not match "
                f"the number of parameter value sets ({len(parameter_values)})."
            )

        for i, value in zip(circuits, parameter_values):
            if len(value) != len(self._parameters[i]):
                raise ValueError(
                    f"The number of values ({len(value)}) does not match "
                    f"the number of parameters ({len(self._parameters[i])}) for the {i}-th circuit."
                )

        for circ_i, obs_i in zip(circuits, observables):
            circuit_num_qubits = self.circuits[circ_i].num_qubits
            observable_num_qubits = self.observables[obs_i].num_qubits
            if circuit_num_qubits != observable_num_qubits:
                raise ValueError(
                    f"The number of qubits of the {circ_i}-th circuit ({circuit_num_qubits}) does "
                    f"not match the number of qubits of the {obs_i}-th observable "
                    f"({observable_num_qubits})."
                )

        if max(circuits) >= len(self.circuits):
            raise ValueError(
                f"The number of circuits is {len(self.circuits)}, "
                f"but the index {max(circuits)} is given."
            )
        if max(observables) >= len(self.observables):
            raise ValueError(
                f"The number of circuits is {len(self.observables)}, "
                f"but the index {max(observables)} is given."
            )
        run_opts = copy(self.options)
        run_opts.update_options(**run_options)

        return self._call(
            circuits=circuits,
            observables=observables,
            parameter_values=parameter_values,
            **run_opts.__dict__,
        )

    @abstractmethod
    def _call(
        self,
        circuits: Sequence[int],
        observables: Sequence[int],
        parameter_values: Sequence[Sequence[float]],
        **run_options,
    ) -> EstimatorResult:
        ...



# This code is part of Qiskit.
#
# (C) Copyright IBM 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Deprecated.

Style sheets for pulse visualization.
"""
import warnings

from collections import namedtuple
from typing import Optional, Tuple

import logging

logger = logging.getLogger(__name__)
ComplexColors = namedtuple("ComplexColors", ["real", "imaginary"])
SchedTableColors = namedtuple("SchedTableColors", ["time", "channel", "event"])


class SchedStyle:
    """Style sheet for Qiskit-Pulse schedule drawer."""

    def __init__(
        self,
        figsize: Optional[Tuple[float, float]] = (10.0, 12.0),
        fig_unit_h_table: float = 0.4,
        use_table: bool = True,
        table_columns: int = 2,
        table_font_size: int = 10,
        axis_font_size: int = 18,
        label_font_size: int = 10,
        icon_font_size: int = 18,
        title_font_size: Optional[int] = 25,
        label_ch_linestyle: str = "--",
        label_ch_color: str = "#222222",
        label_ch_alpha: float = 0.3,
        d_ch_color: ComplexColors = ("#648fff", "#002999"),
        u_ch_color: ComplexColors = ("#ffb000", "#994A00"),
        m_ch_color: ComplexColors = ("#dc267f", "#760019"),
        s_ch_color: str = "#7da781",
        s_ch_linestyle: str = "-",
        table_color: SchedTableColors = ("#e0e0e0", "#f6f6f6", "#f6f6f6"),
        bg_color: str = "#f2f3f4",
        num_points: int = 1000,
        dpi: Optional[int] = 150,
        remove_spacing: bool = True,
        max_table_ratio: float = 0.5,
        vertical_span: float = 0.2,
        axis_formatter: str = "%s",
    ):
        """Deprecated.

        Create new style sheet.

        For any of the Optional fields, if that field is None then it will revert to its
        matplotlib.rcParams counterpart. See for usage on rcParams.  Each argument that
        is optional also describes it's mapped rcParam key below.

        Args:
            figsize: Size of the figure.
                If ``None``, will default to the figure size of the drawing backend.
                If the output is ``matplotlib``, the default
                parameter is ``rcParams['figure.figsize']``.
            fig_unit_h_table: Height of row of event table. See Example.
            use_table: When set `True` use event table.
            table_columns: Number of event table columns.
            table_font_size: Font size of event table.
            axis_font_size: Font size of channel aliases.
                If ``None``, will revert to the axis label size of the drawing backend.
                If the output is ``matplotlib``, the default parameter is
                ``rcParams['axes.titlesize']``.
            label_font_size: Font size of labels in canvas.
            icon_font_size: Size of symbols.
            title_font_size: Font size of schedule name in title.
                If ``None``, will default to the title font size of the drawing backend.
                If the output is ``matplotlib``, the default
                parameter is ``rcParams['figure.titlesize']``.
            label_ch_linestyle: Line style for channel pulse label line.
            label_ch_color: Color code or name of color for channel pulse label line.
            label_ch_alpha: Transparency for channel pulse label line from 0 to 1.
            d_ch_color: Color code or name of colors for real and imaginary part
                of waveform at d channels.
            u_ch_color: Color code or name of colors for real and imaginary part
                of waveform at u channels.
            m_ch_color: Color code or name of colors for real and imaginary part
                of waveform at m channels.
            s_ch_color: Color code or name of color for snapshot channel line.
            s_ch_linestyle: Line style for snapshot line.
            table_color: Color code or name of color for event table columns of
                time, channel, event information.
            bg_color: Color code or name of color for canvas background.
            num_points: Number of points for interpolation of each channel.
            dpi: Resolution in the unit of dot per inch to save image.
                If ``None``, will revert to the DPI setting of the drawing backend.
                If the output is ``matplotlib``, the default
                parameter is ``rcParams['figure.dpi']``.
            remove_spacing: Remove redundant spacing
                when the waveform has no negative values.
            max_table_ratio: Maximum portion of the plot the table can take up.
                Limited to range between 0 and 1.
            vertical_span: Spacing on top and bottom of pulse canvas.
            axis_formatter: Format of horizontal axis of the plot. This is convenient when
                you set ``dt`` option for the drawer. For example, formatter of ``%.3e`` gives you
                horizontal axis values in the scientific notation with 3 digits.

        Example:
            Height of the event table is decided by multiple parameters.::

                figsize = (10.0, 12.0)
                fig_unit_h_table = 0.4
                table_columns = 2
                max_table_ratio = 0.5

            With this setup, events are shown in double-column style with
            each line height of 0.4 inch and the table cannot exceed 5 inch.
            Thus 12 lines are maximum and up to 24 events can be shown.
            If you want to show more events, increase figure height or
            reduce size of line height and table font size.
        """
        warnings.warn(
            "The legacy pulse drawer is deprecated along with this stylesheet. "
            "Please use new drawer `qiskit.visualization.pulse_drawer_v2` "
            "with new stylesheet classes provided by `qiskit.visualization.pulse_v2`. "
            "You can choose one of `IQXStandard`, `IQXSimple`, `IQXDebugging`.",
            DeprecationWarning,
        )

        self.figsize = figsize
        self.fig_unit_h_table = fig_unit_h_table
        self.use_table = use_table
        self.table_columns = table_columns
        self.table_font_size = table_font_size
        self.axis_font_size = axis_font_size
        self.label_font_size = label_font_size
        self.icon_font_size = icon_font_size
        self.title_font_size = title_font_size
        self.d_ch_color = d_ch_color
        self.label_ch_linestyle = label_ch_linestyle
        self.label_ch_color = label_ch_color
        self.label_ch_alpha = label_ch_alpha
        self.u_ch_color = u_ch_color
        self.m_ch_color = m_ch_color
        self.a_ch_color = m_ch_color
        self.s_ch_color = s_ch_color
        self.s_ch_linestyle = s_ch_linestyle
        self.table_color = table_color
        self.bg_color = bg_color
        self.num_points = num_points
        self.dpi = dpi
        self.remove_spacing = remove_spacing

        if max_table_ratio < 0.0 or max_table_ratio > 1.0:
            logger.warning(
                "max_table_ratio of %.2f is not in range [0.0, 1.0], clamping...", max_table_ratio
            )

        self.max_table_ratio = max(min(max_table_ratio, 0.0), 1.0)
        self.vertical_span = vertical_span
        self.axis_formatter = axis_formatter


class PulseStyle:
    """Deprecated.

    Style sheet for Qiskit-Pulse sample pulse drawer."""

    def __init__(
        self,
        figsize: Optional[Tuple[float, float]] = (7.0, 5.0),
        title_font_size: Optional[int] = 18,
        wave_color: ComplexColors = ("#ff0000", "#0000ff"),
        bg_color: str = "#f2f3f4",
        num_points: int = 1000,
        dpi: Optional[int] = None,
    ):
        """Create new style sheet.

        For any of the Optional fields, if that field is None then it will revert to its
        matplotlib.rcParams counterpart. See for usage on rcParams.  Each argument that
        is optional also describes it's mapped rcParam key below.

        Args:
            figsize: Size of the figure.
                If ``None``, will default to the figure size of the drawing backend.
                If the output is ``matplotlib``, the default
                parameter is ``rcParams['figure.figsize']``.
            title_font_size: Font size of schedule name in title.
                If ``None``, will default to the title font size of the drawing backend.
                If the output is ``matplotlib``, the default
                parameter is ``rcParams['figure.titlesize']``.
            wave_color: Color code or name of colors for
                the real and imaginary parts of the waveform.
            bg_color: Color code or name of color for pulse canvas background.
            num_points: Number of points for interpolation.
            dpi: Resolution in the unit of dot per inch to save image.
                If ``None``, will revert to the DPI setting of the drawing backend.
                If the output is ``matplotlib``, the default
                parameter is ``rcParams['figure.dpi']``.
        """
        warnings.warn(
            "The legacy pulse drawer is deprecated along with this stylesheet. "
            "Please use new drawer `qiskit.visualization.pulse_drawer_v2` "
            "with new stylesheet classes provided by `qiskit.visualization.pulse_v2`. "
            "You can choose one of `IQXStandard`, `IQXSimple`, `IQXDebugging`.",
            DeprecationWarning,
        )

        self.figsize = figsize
        self.title_font_size = title_font_size
        self.wave_color = wave_color
        self.bg_color = bg_color
        self.num_points = num_points
        self.dpi = dpi



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Test the BasisTranslator pass"""

from qiskit.transpiler.passes.basis import UnrollCustomDefinitions

from qiskit.test import QiskitTestCase
from qiskit.circuit import EquivalenceLibrary, Gate, Qubit, Clbit
from qiskit.circuit import QuantumCircuit, QuantumRegister
from qiskit.converters import circuit_to_dag
from qiskit.exceptions import QiskitError


class TestGate(Gate):
    """Mock one qubit zero param gate."""

    def __init__(self):
        super().__init__("tg", 1, [])


class TestCompositeGate(Gate):
    """Mock one qubit zero param gate."""

    def __init__(self):
        super().__init__("tcg", 1, [])


class TestUnrollCustomDefinitions(QiskitTestCase):
    """Test the UnrollCustomDefinitions pass."""

    def test_dont_unroll_a_gate_in_eq_lib(self):
        """Verify we don't unroll a gate found in equivalence_library."""
        eq_lib = EquivalenceLibrary()

        gate = TestGate()
        equiv = QuantumCircuit(1)
        equiv.h(0)

        eq_lib.add_equivalence(gate, equiv)

        qc = QuantumCircuit(1)
        qc.append(gate, [0])

        dag = circuit_to_dag(qc)
        out = UnrollCustomDefinitions(eq_lib, ["u3", "cx"]).run(dag)

        expected = qc.copy()
        expected_dag = circuit_to_dag(expected)

        self.assertEqual(out, expected_dag)

    def test_dont_unroll_a_gate_in_basis_gates(self):
        """Verify we don't unroll a gate in basis_gates."""
        eq_lib = EquivalenceLibrary()

        gate = TestGate()
        qc = QuantumCircuit(1)
        qc.append(gate, [0])

        dag = circuit_to_dag(qc)
        out = UnrollCustomDefinitions(eq_lib, ["u3", "cx", "tg"]).run(dag)

        expected = qc.copy()
        expected_dag = circuit_to_dag(expected)

        self.assertEqual(out, expected_dag)

    def test_raise_for_opaque_not_in_eq_lib(self):
        """Verify we raise for an opaque gate not in basis_gates or eq_lib."""
        eq_lib = EquivalenceLibrary()

        gate = TestGate()
        qc = QuantumCircuit(1)
        qc.append(gate, [0])

        dag = circuit_to_dag(qc)
        with self.assertRaisesRegex(QiskitError, "Cannot unroll"):
            UnrollCustomDefinitions(eq_lib, ["u3", "cx"]).run(dag)

    def test_unroll_gate_until_reach_basis_gates(self):
        """Verify we unroll gates until we hit basis_gates."""
        eq_lib = EquivalenceLibrary()

        gate = TestCompositeGate()
        q = QuantumRegister(1, "q")
        gate.definition = QuantumCircuit(q)
        gate.definition.append(TestGate(), [q[0]], [])

        qc = QuantumCircuit(q)
        qc.append(gate, [0])

        dag = circuit_to_dag(qc)
        out = UnrollCustomDefinitions(eq_lib, ["u3", "cx", "tg"]).run(dag)

        expected = QuantumCircuit(1)
        expected.append(TestGate(), [0])
        expected_dag = circuit_to_dag(expected)

        self.assertEqual(out, expected_dag)

    def test_unroll_twice_until_we_get_to_eqlib(self):
        """Verify we unroll gates until we hit basis_gates."""
        eq_lib = EquivalenceLibrary()

        base_gate = TestGate()
        equiv = QuantumCircuit(1)
        equiv.h(0)

        eq_lib.add_equivalence(base_gate, equiv)

        gate = TestCompositeGate()

        q = QuantumRegister(1, "q")
        gate.definition = QuantumCircuit(q)
        gate.definition.append(TestGate(), [q[0]], [])

        qc = QuantumCircuit(1)
        qc.append(gate, [0])

        dag = circuit_to_dag(qc)
        out = UnrollCustomDefinitions(eq_lib, ["u3", "cx"]).run(dag)

        expected = QuantumCircuit(1)
        expected.append(TestGate(), [0])
        expected_dag = circuit_to_dag(expected)

        self.assertEqual(out, expected_dag)

    def test_if_else(self):
        """Test that a simple if-else unrolls correctly."""
        eq_lib = EquivalenceLibrary()

        equiv = QuantumCircuit(1)
        equiv.h(0)
        eq_lib.add_equivalence(TestGate(), equiv)

        equiv = QuantumCircuit(1)
        equiv.z(0)
        eq_lib.add_equivalence(TestCompositeGate(), equiv)

        pass_ = UnrollCustomDefinitions(eq_lib, basis_gates=["h", "z", "cx"])

        true_body = QuantumCircuit(1)
        true_body.h(0)
        true_body.append(TestGate(), [0])
        false_body = QuantumCircuit(1)
        false_body.append(TestCompositeGate(), [0])

        test = QuantumCircuit(1, 1)
        test.h(0)
        test.measure(0, 0)
        test.if_else((0, True), true_body, false_body, [0], [])

        expected = QuantumCircuit(1, 1)
        expected.h(0)
        expected.measure(0, 0)
        expected.if_else((0, True), pass_(true_body), pass_(false_body), [0], [])

        self.assertEqual(pass_(test), expected)

    def test_nested_control_flow(self):
        """Test that the unroller recurses into nested control flow."""
        eq_lib = EquivalenceLibrary()
        base_gate = TestGate()
        equiv = QuantumCircuit(1)
        equiv.h(0)
        eq_lib.add_equivalence(base_gate, equiv)
        base_gate = TestCompositeGate()
        equiv = QuantumCircuit(1)
        equiv.z(0)
        eq_lib.add_equivalence(base_gate, equiv)

        pass_ = UnrollCustomDefinitions(eq_lib, basis_gates=["h", "z", "cx"])

        qubit = Qubit()
        clbit = Clbit()

        for_body = QuantumCircuit(1)
        for_body.append(TestGate(), [0], [])

        while_body = QuantumCircuit(1)
        while_body.append(TestCompositeGate(), [0], [])

        true_body = QuantumCircuit([qubit, clbit])
        true_body.while_loop((clbit, True), while_body, [0], [0])

        test = QuantumCircuit([qubit, clbit])
        test.for_loop(range(2), None, for_body, [0], [0])
        test.if_else((clbit, True), true_body, None, [0], [0])

        expected_if_body = QuantumCircuit([qubit, clbit])
        expected_if_body.while_loop((clbit, True), pass_(while_body), [0], [0])
        expected = QuantumCircuit([qubit, clbit])
        expected.for_loop(range(2), None, pass_(for_body), [0], [0])
        expected.if_else(range(2), pass_(expected_if_body), None, [0], [0])

        self.assertEqual(pass_(test), expected)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Quantum measurement in the computational basis.
"""

import warnings

from qiskit.circuit.instruction import Instruction
from qiskit.circuit.exceptions import CircuitError


class Measure(Instruction):
    """Quantum measurement in the computational basis."""

    def __init__(self):
        """Create new measurement instruction."""
        super().__init__("measure", 1, 1, [])

    def broadcast_arguments(self, qargs, cargs):
        qarg = qargs[0]
        carg = cargs[0]

        if len(carg) == len(qarg):
            for qarg, carg in zip(qarg, carg):
                yield [qarg], [carg]
        elif len(qarg) == 1 and carg:
            for each_carg in carg:
                yield qarg, [each_carg]
        else:
            raise CircuitError("register size error")


def measure(circuit, qubit, clbit):
    """Measure a quantum bit into classical bit.

    .. deprecated:: Qiskit Terra 0.19
        Use :meth:`.QuantumCircuit.measure` instead, either by calling
        ``circuit.measure(qubit, clbit)``, or if a full function is required, then
        ``QuantumCircuit.measure(circuit, qubit, clbit)``.

    Args:
        circuit (QuantumCircuit): the quantum circuit to attach the measurement to.
        qubit (Union[Qubit, int]): the quantum bit to measure
        clbit (Union[Clbit, int]): the classical bit to store the measurement result in.

    Returns:
        .InstructionSet: a handle to the created instruction.

    Raises:
        CircuitError: if either bit is not in the circuit, or is in a bad format.
    """
    warnings.warn(
        "The loose 'measure' function is deprecated as of Qiskit Terra 0.19, and will be removed"
        " in a future release.  Instead, you should call 'circuit.measure(qubit, clbit)', or if you"
        " need a function, you can do `QuantumCircuit.measure(circuit, qubit, clbit)'.",
        category=DeprecationWarning,
        stacklevel=2,
    )
    return circuit.measure(qubit, clbit)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Module for common pulse programming utilities."""
import functools
from typing import List, Dict, Union

import numpy as np

from qiskit.circuit.parameterexpression import ParameterExpression
from qiskit.pulse.exceptions import UnassignedDurationError, QiskitError
from qiskit.utils import deprecate_function  # pylint: disable=cyclic-import


def format_meas_map(meas_map: List[List[int]]) -> Dict[int, List[int]]:
    """
    Return a mapping from qubit label to measurement group given the nested list meas_map returned
    by a backend configuration. (Qubits can not always be measured independently.) Sorts the
    measurement group for consistency.

    Args:
        meas_map: Groups of qubits that get measured together, for example: [[0, 1], [2, 3, 4]]
    Returns:
        Measure map in map format
    """
    qubit_mapping = {}
    for sublist in meas_map:
        sublist.sort()
        for q in sublist:
            qubit_mapping[q] = sublist
    return qubit_mapping


@functools.lru_cache(maxsize=None)
def format_parameter_value(
    operand: ParameterExpression,
    decimal: int = 10,
) -> Union[ParameterExpression, complex]:
    """Convert ParameterExpression into the most suitable data type.

    Args:
        operand: Operand value in arbitrary data type including ParameterExpression.
        decimal: Number of digit to round returned value.

    Returns:
        Value casted to non-parameter data type, when possible.
    """
    try:
        # value is assigned.
        # note that ParameterExpression directly supports __complex__ via sympy or symengine
        evaluated = complex(operand)
        # remove truncation error
        evaluated = np.round(evaluated, decimals=decimal)
        # typecast into most likely data type
        if np.isreal(evaluated):
            evaluated = float(evaluated.real)
            if evaluated.is_integer():
                evaluated = int(evaluated)

        return evaluated
    except TypeError:
        # value is not assigned
        pass

    return operand


def instruction_duration_validation(duration: int):
    """Validate instruction duration.

    Args:
        duration: Instruction duration value to validate.

    Raises:
        UnassignedDurationError: When duration is unassigned.
        QiskitError: When invalid duration is assigned.
    """
    if isinstance(duration, ParameterExpression):
        raise UnassignedDurationError(
            "Instruction duration {} is not assigned. "
            "Please bind all durations to an integer value before playing in the Schedule, "
            "or use ScheduleBlock to align instructions with unassigned duration."
            "".format(repr(duration))
        )

    if not isinstance(duration, (int, np.integer)) or duration < 0:
        raise QiskitError(
            f"Instruction duration must be a non-negative integer, got {duration} instead."
        )


@deprecate_function("Deprecated since Terra 0.22.0. Use 'qiskit.utils.deprecate_function' instead.")
def deprecated_functionality(func):
    """A decorator that raises deprecation warning without showing alternative method."""
    return deprecate_function(
        f"Calling {func.__name__} is being deprecated and will be removed soon. "
        "No alternative method will be provided with this change. "
        "If there is any practical usage of this functionality, please write "
        "an issue in Qiskit/qiskit-terra repository.",
        category=DeprecationWarning,
        stacklevel=2,
    )(func)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Padding pass to fill empty timeslot."""

from typing import List, Optional, Union

from qiskit.circuit import Qubit, Clbit, Instruction
from qiskit.circuit.delay import Delay
from qiskit.dagcircuit import DAGCircuit, DAGNode
from qiskit.transpiler.basepasses import TransformationPass
from qiskit.transpiler.exceptions import TranspilerError


class BasePadding(TransformationPass):
    """The base class of padding pass.

    This pass requires one of scheduling passes to be executed before itself.
    Since there are multiple scheduling strategies, the selection of scheduling
    pass is left in the hands of the pass manager designer.
    Once a scheduling analysis pass is run, ``node_start_time`` is generated
    in the :attr:`property_set`.  This information is represented by a python dictionary of
    the expected instruction execution times keyed on the node instances.
    Entries in the dictionary are only created for non-delay nodes.
    The padding pass expects all ``DAGOpNode`` in the circuit to be scheduled.

    This base class doesn't define any sequence to interleave, but it manages
    the location where the sequence is inserted, and provides a set of information necessary
    to construct the proper sequence. Thus, a subclass of this pass just needs to implement
    :meth:`_pad` method, in which the subclass constructs a circuit block to insert.
    This mechanism removes lots of boilerplate logic to manage whole DAG circuits.

    Note that padding pass subclasses should define interleaving sequences satisfying:

        - Interleaved sequence does not change start time of other nodes
        - Interleaved sequence should have total duration of the provided ``time_interval``.

    Any manipulation violating these constraints may prevent this base pass from correctly
    tracking the start time of each instruction,
    which may result in violation of hardware alignment constraints.
    """

    def run(self, dag: DAGCircuit):
        """Run the padding pass on ``dag``.

        Args:
            dag: DAG to be checked.

        Returns:
            DAGCircuit: DAG with idle time filled with instructions.

        Raises:
            TranspilerError: When a particular node is not scheduled, likely some transform pass
                is inserted before this node is called.
        """
        self._pre_runhook(dag)

        node_start_time = self.property_set["node_start_time"].copy()

        new_dag = DAGCircuit()
        for qreg in dag.qregs.values():
            new_dag.add_qreg(qreg)
        for creg in dag.cregs.values():
            new_dag.add_creg(creg)

        # Update start time dictionary for the new_dag.
        # This information may be used for further scheduling tasks,
        # but this is immediately invalidated becasue node id is updated in the new_dag.
        self.property_set["node_start_time"].clear()

        new_dag.name = dag.name
        new_dag.metadata = dag.metadata
        new_dag.unit = self.property_set["time_unit"]
        new_dag.calibrations = dag.calibrations
        new_dag.global_phase = dag.global_phase

        idle_after = {bit: 0 for bit in dag.qubits}

        # Compute fresh circuit duration from the node start time dictionary and op duration.
        # Note that pre-scheduled duration may change within the alignment passes, i.e.
        # if some instruction time t0 violating the hardware alignment constraint,
        # the alignment pass may delay t0 and accordingly the circuit duration changes.
        circuit_duration = 0
        for node in dag.topological_op_nodes():
            if node in node_start_time:
                t0 = node_start_time[node]
                t1 = t0 + node.op.duration
                circuit_duration = max(circuit_duration, t1)

                if isinstance(node.op, Delay):
                    # The padding class considers a delay instruction as idle time
                    # rather than instruction. Delay node is removed so that
                    # we can extract non-delay predecessors.
                    dag.remove_op_node(node)
                    continue

                for bit in node.qargs:

                    # Fill idle time with some sequence
                    if t0 - idle_after[bit] > 0:
                        # Find previous node on the wire, i.e. always the latest node on the wire
                        prev_node = next(new_dag.predecessors(new_dag.output_map[bit]))
                        self._pad(
                            dag=new_dag,
                            qubit=bit,
                            t_start=idle_after[bit],
                            t_end=t0,
                            next_node=node,
                            prev_node=prev_node,
                        )

                    idle_after[bit] = t1

                self._apply_scheduled_op(new_dag, t0, node.op, node.qargs, node.cargs)
            else:
                raise TranspilerError(
                    f"Operation {repr(node)} is likely added after the circuit is scheduled. "
                    "Schedule the circuit again if you transformed it."
                )

        # Add delays until the end of circuit.
        for bit in new_dag.qubits:
            if circuit_duration - idle_after[bit] > 0:
                node = new_dag.output_map[bit]
                prev_node = next(new_dag.predecessors(node))
                self._pad(
                    dag=new_dag,
                    qubit=bit,
                    t_start=idle_after[bit],
                    t_end=circuit_duration,
                    next_node=node,
                    prev_node=prev_node,
                )

        new_dag.duration = circuit_duration

        return new_dag

    def _pre_runhook(self, dag: DAGCircuit):
        """Extra routine inserted before running the padding pass.

        Args:
            dag: DAG circuit on which the sequence is applied.

        Raises:
            TranspilerError: If the whole circuit or instruction is not scheduled.
        """
        if "node_start_time" not in self.property_set:
            raise TranspilerError(
                f"The input circuit {dag.name} is not scheduled. Call one of scheduling passes "
                f"before running the {self.__class__.__name__} pass."
            )

    def _apply_scheduled_op(
        self,
        dag: DAGCircuit,
        t_start: int,
        oper: Instruction,
        qubits: Union[Qubit, List[Qubit]],
        clbits: Optional[Union[Clbit, List[Clbit]]] = None,
    ):
        """Add new operation to DAG with scheduled information.

        This is identical to apply_operation_back + updating the node_start_time propety.

        Args:
            dag: DAG circuit on which the sequence is applied.
            t_start: Start time of new node.
            oper: New operation that is added to the DAG circuit.
            qubits: The list of qubits that the operation acts on.
            clbits: The list of clbits that the operation acts on.
        """
        if isinstance(qubits, Qubit):
            qubits = [qubits]
        if isinstance(clbits, Clbit):
            clbits = [clbits]

        new_node = dag.apply_operation_back(oper, qargs=qubits, cargs=clbits)
        self.property_set["node_start_time"][new_node] = t_start

    def _pad(
        self,
        dag: DAGCircuit,
        qubit: Qubit,
        t_start: int,
        t_end: int,
        next_node: DAGNode,
        prev_node: DAGNode,
    ):
        """Interleave instruction sequence in between two nodes.

        .. note::
            If a DAGOpNode is added here, it should update node_start_time property
            in the property set so that the added node is also scheduled.
            This is achieved by adding operation via :meth:`_apply_scheduled_op`.

        .. note::

            This method doesn't check if the total duration of new DAGOpNode added here
            is identical to the interval (``t_end - t_start``).
            A developer of the pass must guarantee this is satisfied.
            If the duration is greater than the interval, your circuit may be
            compiled down to the target code with extra duration on the backend compiler,
            which is then played normally without error. However, the outcome of your circuit
            might be unexpected due to erroneous scheduling.

        Args:
            dag: DAG circuit that sequence is applied.
            qubit: The wire that the sequence is applied on.
            t_start: Absolute start time of this interval.
            t_end: Absolute end time of this interval.
            next_node: Node that follows the sequence.
            prev_node: Node ahead of the sequence.
        """
        raise NotImplementedError



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Node for an OPENQASM file identifier/version statement."""

import re

from .node import Node


class Format(Node):
    """Node for an OPENQASM file identifier/version statement."""

    def __init__(self, value):
        """Create the version node."""
        super().__init__("format", None, None)
        parts = re.match(r"(\w+)\s+(\d+)(\.(\d+))?", value)
        self.language = parts.group(1)
        self.majorversion = parts.group(2)
        self.minorversion = parts.group(4) if parts.group(4) is not None else "0"

    def version(self):
        """Return the version."""
        return f"{self.majorversion}.{self.minorversion}"

    def qasm(self):
        """Return the corresponding format string."""
        return f"{self.language} {self.version()};"



# This code is part of Qiskit.
#
# (C) Copyright IBM 2018, 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

""" Minimize using objective function """

from typing import List, Optional, Tuple, Callable
from enum import Enum
from abc import abstractmethod
import logging
import numpy as np

from qiskit.utils import optionals as _optionals
from ..optimizer import Optimizer, OptimizerSupportLevel, OptimizerResult, POINT

logger = logging.getLogger(__name__)


class NLoptOptimizerType(Enum):
    """NLopt Valid Optimizer"""

    GN_CRS2_LM = 1
    GN_DIRECT_L_RAND = 2
    GN_DIRECT_L = 3
    GN_ESCH = 4
    GN_ISRES = 5


@_optionals.HAS_NLOPT.require_in_instance
class NLoptOptimizer(Optimizer):
    """
    NLopt global optimizer base class
    """

    # pylint: disable=import-error

    _OPTIONS = ["max_evals"]

    def __init__(self, max_evals: int = 1000) -> None:  # pylint: disable=unused-argument
        """
        Args:
            max_evals: Maximum allowed number of function evaluations.

        Raises:
            MissingOptionalLibraryError: NLopt library not installed.
        """
        import nlopt

        super().__init__()
        for k, v in list(locals().items()):
            if k in self._OPTIONS:
                self._options[k] = v

        self._optimizer_names = {
            NLoptOptimizerType.GN_CRS2_LM: nlopt.GN_CRS2_LM,
            NLoptOptimizerType.GN_DIRECT_L_RAND: nlopt.GN_DIRECT_L_RAND,
            NLoptOptimizerType.GN_DIRECT_L: nlopt.GN_DIRECT_L,
            NLoptOptimizerType.GN_ESCH: nlopt.GN_ESCH,
            NLoptOptimizerType.GN_ISRES: nlopt.GN_ISRES,
        }

    @abstractmethod
    def get_nlopt_optimizer(self) -> NLoptOptimizerType:
        """return NLopt optimizer enum type"""
        raise NotImplementedError

    def get_support_level(self):
        """return support level dictionary"""
        return {
            "gradient": OptimizerSupportLevel.ignored,
            "bounds": OptimizerSupportLevel.supported,
            "initial_point": OptimizerSupportLevel.required,
        }

    @property
    def settings(self):
        return {"max_evals": self._options.get("max_evals", 1000)}

    def minimize(
        self,
        fun: Callable[[POINT], float],
        x0: POINT,
        jac: Optional[Callable[[POINT], POINT]] = None,
        bounds: Optional[List[Tuple[float, float]]] = None,
    ) -> OptimizerResult:
        import nlopt

        x0 = np.asarray(x0)

        if bounds is None:
            bounds = [(None, None)] * x0.size

        threshold = 3 * np.pi
        low = [(l if l is not None else -threshold) for (l, u) in bounds]
        high = [(u if u is not None else threshold) for (l, u) in bounds]

        name = self._optimizer_names[self.get_nlopt_optimizer()]
        opt = nlopt.opt(name, len(low))
        logger.debug(opt.get_algorithm_name())

        opt.set_lower_bounds(low)
        opt.set_upper_bounds(high)

        eval_count = 0

        def wrap_objfunc_global(x, _grad):
            nonlocal eval_count
            eval_count += 1
            return fun(x)

        opt.set_min_objective(wrap_objfunc_global)
        opt.set_maxeval(self._options.get("max_evals", 1000))

        xopt = opt.optimize(x0)
        minf = opt.last_optimum_value()

        logger.debug("Global minimize found %s eval count %s", minf, eval_count)

        result = OptimizerResult()
        result.x = xopt
        result.fun = minf
        result.nfev = eval_count

        return result



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2018.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Helper functions for building dictionaries from matrices and lists."""

import numpy as np


def make_dict_observable(matrix_observable):
    """Convert an observable in matrix form to dictionary form.

    Takes in a diagonal observable as a matrix and converts it to a dictionary
    form. Can also handle a list sorted of the diagonal elements.

    Args:
        matrix_observable (list): The observable to be converted to dictionary
        form. Can be a matrix or just an ordered list of observed values

    Returns:
        Dict: A dictionary with all observable states as keys, and corresponding
        values being the observed value for that state
    """
    dict_observable = {}
    observable = np.array(matrix_observable)
    observable_size = len(observable)
    observable_bits = int(np.ceil(np.log2(observable_size)))
    binary_formatter = f"0{observable_bits}b"
    if observable.ndim == 2:
        observable = observable.diagonal()
    for state_no in range(observable_size):
        state_str = format(state_no, binary_formatter)
        dict_observable[state_str] = observable[state_no]
    return dict_observable



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.


"""
Pauli Transfer Matrix (PTM) representation of a Quantum Channel.
"""

import copy
import numpy as np

from qiskit.circuit.quantumcircuit import QuantumCircuit
from qiskit.circuit.instruction import Instruction
from qiskit.exceptions import QiskitError
from qiskit.quantum_info.operators.channel.quantum_channel import QuantumChannel
from qiskit.quantum_info.operators.channel.superop import SuperOp
from qiskit.quantum_info.operators.channel.transformations import _to_ptm
from qiskit.quantum_info.operators.mixins import generate_apidocs


class PTM(QuantumChannel):
    r"""Pauli Transfer Matrix (PTM) representation of a Quantum Channel.

    The PTM representation of an :math:`n`-qubit quantum channel
    :math:`\mathcal{E}` is an :math:`n`-qubit :class:`SuperOp` :math:`R`
    defined with respect to vectorization in the Pauli basis instead of
    column-vectorization. The elements of the PTM :math:`R` are
    given by

    .. math::

        R_{i,j} = \mbox{Tr}\left[P_i \mathcal{E}(P_j) \right]

    where :math:`[P_0, P_1, ..., P_{4^{n}-1}]` is the :math:`n`-qubit Pauli basis in
    lexicographic order.

    Evolution of a :class:`~qiskit.quantum_info.DensityMatrix`
    :math:`\rho` with respect to the PTM is given by

    .. math::

        |\mathcal{E}(\rho)\rangle\!\rangle_P = S_P |\rho\rangle\!\rangle_P

    where :math:`|A\rangle\!\rangle_P` denotes vectorization in the Pauli basis
    :math:`\langle i | A\rangle\!\rangle_P = \mbox{Tr}[P_i A]`.

    See reference [1] for further details.

    References:
        1. C.J. Wood, J.D. Biamonte, D.G. Cory, *Tensor networks and graphical calculus
           for open quantum systems*, Quant. Inf. Comp. 15, 0579-0811 (2015).
           `arXiv:1111.6950 [quant-ph] <https://arxiv.org/abs/1111.6950>`_
    """

    def __init__(self, data, input_dims=None, output_dims=None):
        """Initialize a PTM quantum channel operator.

        Args:
            data (QuantumCircuit or
                  Instruction or
                  BaseOperator or
                  matrix): data to initialize superoperator.
            input_dims (tuple): the input subsystem dimensions.
                                [Default: None]
            output_dims (tuple): the output subsystem dimensions.
                                 [Default: None]

        Raises:
            QiskitError: if input data is not an N-qubit channel or
                         cannot be initialized as a PTM.

        Additional Information:
            If the input or output dimensions are None, they will be
            automatically determined from the input data. The PTM
            representation is only valid for N-qubit channels.
        """
        # If the input is a raw list or matrix we assume that it is
        # already a Chi matrix.
        if isinstance(data, (list, np.ndarray)):
            # Should we force this to be real?
            ptm = np.asarray(data, dtype=complex)
            # Determine input and output dimensions
            dout, din = ptm.shape
            if input_dims:
                input_dim = np.product(input_dims)
            else:
                input_dim = int(np.sqrt(din))
            if output_dims:
                output_dim = np.product(input_dims)
            else:
                output_dim = int(np.sqrt(dout))
            if output_dim**2 != dout or input_dim**2 != din or input_dim != output_dim:
                raise QiskitError("Invalid shape for PTM matrix.")
        else:
            # Otherwise we initialize by conversion from another Qiskit
            # object into the QuantumChannel.
            if isinstance(data, (QuantumCircuit, Instruction)):
                # If the input is a Terra QuantumCircuit or Instruction we
                # convert it to a SuperOp
                data = SuperOp._init_instruction(data)
            else:
                # We use the QuantumChannel init transform to initialize
                # other objects into a QuantumChannel or Operator object.
                data = self._init_transformer(data)
            input_dim, output_dim = data.dim
            # Now that the input is an operator we convert it to a PTM object
            rep = getattr(data, "_channel_rep", "Operator")
            ptm = _to_ptm(rep, data._data, input_dim, output_dim)
            if input_dims is None:
                input_dims = data.input_dims()
            if output_dims is None:
                output_dims = data.output_dims()
        # Check input is N-qubit channel
        num_qubits = int(np.log2(input_dim))
        if 2**num_qubits != input_dim or input_dim != output_dim:
            raise QiskitError("Input is not an n-qubit Pauli transfer matrix.")
        super().__init__(ptm, num_qubits=num_qubits)

    def __array__(self, dtype=None):
        if dtype:
            np.asarray(self.data, dtype=dtype)
        return self.data

    @property
    def _bipartite_shape(self):
        """Return the shape for bipartite matrix"""
        return (self._output_dim, self._output_dim, self._input_dim, self._input_dim)

    def _evolve(self, state, qargs=None):
        return SuperOp(self)._evolve(state, qargs)

    # ---------------------------------------------------------------------
    # BaseOperator methods
    # ---------------------------------------------------------------------

    def conjugate(self):
        # Since conjugation is basis dependent we transform
        # to the SuperOp representation to compute the
        # conjugate channel
        return PTM(SuperOp(self).conjugate())

    def transpose(self):
        return PTM(SuperOp(self).transpose())

    def adjoint(self):
        return PTM(SuperOp(self).adjoint())

    def compose(self, other, qargs=None, front=False):
        if qargs is None:
            qargs = getattr(other, "qargs", None)
        if qargs is not None:
            return PTM(SuperOp(self).compose(other, qargs=qargs, front=front))

        # Convert other to PTM
        if not isinstance(other, PTM):
            other = PTM(other)
        new_shape = self._op_shape.compose(other._op_shape, qargs, front)
        input_dims = new_shape.dims_r()
        output_dims = new_shape.dims_l()
        if front:
            data = np.dot(self._data, other.data)
        else:
            data = np.dot(other.data, self._data)
        ret = PTM(data, input_dims, output_dims)
        ret._op_shape = new_shape
        return ret

    def tensor(self, other):
        if not isinstance(other, PTM):
            other = PTM(other)
        return self._tensor(self, other)

    def expand(self, other):
        if not isinstance(other, PTM):
            other = PTM(other)
        return self._tensor(other, self)

    @classmethod
    def _tensor(cls, a, b):
        ret = copy.copy(a)
        ret._op_shape = a._op_shape.tensor(b._op_shape)
        ret._data = np.kron(a._data, b.data)
        return ret


# Update docstrings for API docs
generate_apidocs(PTM)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Unrolls instructions with custom definitions."""

from qiskit.exceptions import QiskitError
from qiskit.transpiler.basepasses import TransformationPass
from qiskit.transpiler.passes.utils import control_flow
from qiskit.circuit import ControlledGate, ControlFlowOp
from qiskit.converters.circuit_to_dag import circuit_to_dag


class UnrollCustomDefinitions(TransformationPass):
    """Unrolls instructions with custom definitions."""

    def __init__(self, equivalence_library, basis_gates):
        """Unrolls instructions with custom definitions.

        Args:
            equivalence_library (EquivalenceLibrary): The equivalence library
                which will be used by the BasisTranslator pass. (Instructions in
                this library will not be unrolled by this pass.)
            basis_gates (list[str]): Target basis names to unroll to, e.g. `['u3', 'cx']`.
        """

        super().__init__()
        self._equiv_lib = equivalence_library
        self._basis_gates = basis_gates

    def run(self, dag):
        """Run the UnrollCustomDefinitions pass on `dag`.

        Args:
            dag (DAGCircuit): input dag

        Raises:
            QiskitError: if unable to unroll given the basis due to undefined
            decomposition rules (such as a bad basis) or excessive recursion.

        Returns:
            DAGCircuit: output unrolled dag
        """

        if self._basis_gates is None:
            return dag

        basic_insts = {"measure", "reset", "barrier", "snapshot", "delay"}
        device_insts = basic_insts | set(self._basis_gates)

        for node in dag.op_nodes():
            if isinstance(node.op, ControlFlowOp):
                node.op = control_flow.map_blocks(self.run, node.op)
                continue

            if getattr(node.op, "_directive", False):
                continue

            if dag.has_calibration_for(node):
                continue

            if node.name in device_insts or self._equiv_lib.has_entry(node.op):
                if isinstance(node.op, ControlledGate) and node.op._open_ctrl:
                    pass
                else:
                    continue

            try:
                rule = node.op.definition.data
            except TypeError as err:
                raise QiskitError(f"Error decomposing node {node.name}: {err}") from err
            except AttributeError:
                # definition is None
                rule = None

            if not rule:
                if rule == []:
                    dag.remove_op_node(node)
                    continue

                # opaque node
                raise QiskitError(
                    "Cannot unroll the circuit to the given basis, %s. "
                    "Instruction %s not found in equivalence library "
                    "and no rule found to expand." % (str(self._basis_gates), node.op.name)
                )
            decomposition = circuit_to_dag(node.op.definition)
            unrolled_dag = UnrollCustomDefinitions(self._equiv_lib, self._basis_gates).run(
                decomposition
            )
            dag.substitute_node_with_dag(node, unrolled_dag)

        return dag



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Combine consecutive Cliffords over the same qubits."""

from qiskit.transpiler.basepasses import TransformationPass
from qiskit.transpiler.passes.utils import control_flow
from qiskit.quantum_info.operators import Clifford


class OptimizeCliffords(TransformationPass):
    """Combine consecutive Cliffords over the same qubits.
    This serves as an example of extra capabilities enabled by storing
    Cliffords natively on the circuit.
    """

    @control_flow.trivial_recurse
    def run(self, dag):
        """Run the OptimizeCliffords pass on `dag`.

        Args:
            dag (DAGCircuit): the DAG to be optimized.

        Returns:
            DAGCircuit: the optimized DAG.
        """

        blocks = []
        prev_node = None
        cur_block = []

        # Iterate over all nodes and collect consecutive Cliffords over the
        # same qubits. In this very first proof-of-concept implementation
        # we require the same ordering of qubits, but this restriction will
        # be shortly removed. An interesting question is whether we may also
        # want to compose Cliffords over different sets of qubits, such as
        # cliff1 over qubits [1, 2, 3] and cliff2 over [2, 3, 4].
        for node in dag.topological_op_nodes():
            if isinstance(node.op, Clifford):
                if prev_node is None:
                    blocks.append(cur_block)
                    cur_block = [node]
                else:
                    if prev_node.qargs == node.qargs:
                        cur_block.append(node)
                    else:
                        blocks.append(cur_block)
                        cur_block = [node]

                prev_node = node

            else:
                # not a clifford
                if cur_block:
                    blocks.append(cur_block)
                prev_node = None
                cur_block = []

        if cur_block:
            blocks.append(cur_block)

        # Replace every discovered block of cliffords by a single clifford
        # based on the Cliffords' compose function.
        for cur_nodes in blocks:
            # Create clifford functions only out of blocks with at least 2 gates
            if len(cur_nodes) <= 1:
                continue

            wire_pos_map = dict((qb, ix) for ix, qb in enumerate(cur_nodes[0].qargs))

            # Construct a linear circuit
            cliff = cur_nodes[0].op
            for i, node in enumerate(cur_nodes):
                if i > 0:
                    cliff = Clifford.compose(node.op, cliff, front=True)

            # Replace the block by the composed clifford
            dag.replace_block_with_op(cur_nodes, cliff, wire_pos_map, cycle_check=False)

        return dag



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Tests preset pass manager with faulty backends"""

from ddt import ddt, data

from qiskit import QuantumCircuit, QuantumRegister, BasicAer, execute
from qiskit.compiler import transpile
from qiskit.test import QiskitTestCase
from qiskit.converters import circuit_to_dag
from qiskit.circuit.library import CXGate
from qiskit.dagcircuit import DAGOpNode
from qiskit.transpiler import TranspilerError
from ..providers.faulty_backends import (
    FakeOurenseFaultyQ1,
    FakeOurenseFaultyCX13,
    FakeOurenseFaultyCX01CX10,
    FakeOurenseFaultyCX13CX31,
)


class TestFaultyBackendCase(QiskitTestCase):
    """Base TestCase for testing transpilation if faulty backends."""

    def assertEqualCount(self, circuit1, circuit2):
        """Asserts circuit1 and circuit2 has the same result counts after execution in BasicAer"""
        backend = BasicAer.get_backend("qasm_simulator")
        shots = 2048

        result1 = (
            execute(
                circuit1,
                backend,
                basis_gates=["u1", "u2", "u3", "id", "cx"],
                seed_simulator=0,
                seed_transpiler=0,
                shots=shots,
            )
            .result()
            .get_counts()
        )

        result2 = (
            execute(
                circuit2,
                backend,
                basis_gates=["u1", "u2", "u3", "id", "cx"],
                seed_simulator=0,
                seed_transpiler=0,
                shots=shots,
            )
            .result()
            .get_counts()
        )

        for key in set(result1.keys()).union(result2.keys()):
            with self.subTest(key=key):
                diff = abs(result1.get(key, 0) - result2.get(key, 0))
                self.assertLess(diff / shots * 100, 2.5)


@ddt
class TestFaultyCX01CX10(TestFaultyBackendCase):
    """Test preset passmanagers with FakeOurenseFaultyCX01CX10
    A fake 5 qubit backend, with a faulty CX(Q0, Q1) (and symmetric).
         0 (↔) 1 ↔ 3 ↔ 4
               ↕
               2
    """

    def assertIdleCX01(self, circuit):
        """Asserts the CX(0, 1) (and symmetric) is not used in the circuit"""
        physical_qubits = QuantumRegister(5, "q")
        cx_nodes = circuit_to_dag(circuit).op_nodes(CXGate)
        for node in cx_nodes:
            if set(node.qargs) == {physical_qubits[0], physical_qubits[1]}:
                raise AssertionError("Faulty CX(Q0, Q1) (or symmetric) is being used.")

    @data(0, 1, 2, 3)
    def test_level(self, level):
        """Test level {level} Ourense backend with faulty CX(Q0, Q1) and CX(Q1, Q0)"""
        circuit = QuantumCircuit(QuantumRegister(4, "qr"))
        circuit.h(range(4))
        circuit.ccx(0, 1, 2)
        circuit.measure_all()
        result = transpile(
            circuit,
            backend=FakeOurenseFaultyCX01CX10(),
            optimization_level=level,
            seed_transpiler=42,
        )

        self.assertIdleCX01(result)
        self.assertEqualCount(circuit, result)

    @data(0, 1, 2, 3)
    def test_layout_level(self, level):
        """Test level {level} with a faulty CX(Q0, Q1) with a working initial layout"""
        circuit = QuantumCircuit(QuantumRegister(4, "qr"))
        circuit.h(range(4))
        circuit.ccx(0, 1, 2)
        circuit.measure_all()
        result = transpile(
            circuit,
            backend=FakeOurenseFaultyCX01CX10(),
            optimization_level=level,
            initial_layout=[3, 4, 1, 2],
            seed_transpiler=42,
        )

        self.assertIdleCX01(result)
        self.assertEqualCount(circuit, result)

    @data(0, 1, 2, 3)
    def test_failing_layout_level(self, level):
        """Test level {level} with a faulty CX(Q0, Q1) with a failing initial layout. Raises."""
        circuit = QuantumCircuit(QuantumRegister(4, "qr"))
        circuit.h(range(4))
        circuit.ccx(0, 1, 2)
        circuit.measure_all()

        message = "The initial_layout parameter refers to faulty or disconnected qubits"

        with self.assertRaises(TranspilerError) as context:
            transpile(
                circuit,
                backend=FakeOurenseFaultyCX01CX10(),
                optimization_level=level,
                initial_layout=[0, 4, 1, 2],
                seed_transpiler=42,
            )

        self.assertEqual(context.exception.message, message)


@ddt
class TestFaultyCX13CX31(TestFaultyBackendCase):
    """Test preset passmanagers with FakeOurenseFaultyCX13CX31
    A fake 5 qubit backend, with a faulty CX(Q1, Q3) (and symmetric).
         0 ↔ 1 (↔) 3 ↔ 4
             ↕
             2
    """

    def assertIdleCX13(self, circuit):
        """Asserts the CX(1, 3) (and symmetric) is not used in the circuit"""
        physical_qubits = QuantumRegister(5, "q")
        cx_nodes = circuit_to_dag(circuit).op_nodes(CXGate)
        for node in cx_nodes:
            if set(node.qargs) == {physical_qubits[1], physical_qubits[3]}:
                raise AssertionError("Faulty CX(Q1, Q3) (or symmetric) is being used.")

    @data(0, 1, 2, 3)
    def test_level(self, level):
        """Test level {level} Ourense backend with faulty CX(Q1, Q3) and CX(Q3, Q1)"""
        circuit = QuantumCircuit(QuantumRegister(3, "qr"))
        circuit.h(range(3))
        circuit.ccx(0, 1, 2)
        circuit.measure_all()
        result = transpile(
            circuit,
            backend=FakeOurenseFaultyCX13CX31(),
            optimization_level=level,
            seed_transpiler=42,
        )

        self.assertIdleCX13(result)
        self.assertEqualCount(circuit, result)

    @data(0, 1, 2, 3)
    def test_layout_level(self, level):
        """Test level {level} with a faulty CX(Q1, Q3) with a working initial layout"""
        circuit = QuantumCircuit(QuantumRegister(3, "qr"))
        circuit.h(range(3))
        circuit.ccx(0, 1, 2)
        circuit.measure_all()
        result = transpile(
            circuit,
            backend=FakeOurenseFaultyCX13CX31(),
            optimization_level=level,
            initial_layout=[0, 2, 1],
            seed_transpiler=42,
        )

        self.assertIdleCX13(result)
        self.assertEqualCount(circuit, result)

    @data(0, 1, 2, 3)
    def test_failing_layout_level(self, level):
        """Test level {level} with a faulty CX(Q1, Q3) with a failing initial layout. Raises."""
        circuit = QuantumCircuit(QuantumRegister(3, "qr"))
        circuit.h(range(3))
        circuit.ccx(0, 1, 2)
        circuit.measure_all()

        message = "The initial_layout parameter refers to faulty or disconnected qubits"

        with self.assertRaises(TranspilerError) as context:
            transpile(
                circuit,
                backend=FakeOurenseFaultyCX13CX31(),
                optimization_level=level,
                initial_layout=[0, 1, 3],
                seed_transpiler=42,
            )

        self.assertEqual(context.exception.message, message)


@ddt
class TestFaultyCX13(TestFaultyBackendCase):
    """Test preset passmanagers with FakeOurenseFaultyCX13
    A fake 5 qubit backend, with a faulty CX(Q1, Q3). Notice that CX(Q3, Q1) is operational
         0 ↔ 1 <- 3 ↔ 4
             ↕
             2
    """

    def assertIdleCX13(self, circuit):
        """Asserts the CX(1, 3) is not used in the circuit"""
        physical_qubits = QuantumRegister(5, "q")
        cx_nodes = circuit_to_dag(circuit).op_nodes(CXGate)
        for node in cx_nodes:
            if node.qargs == [physical_qubits[1], physical_qubits[3]]:
                raise AssertionError("Faulty CX(Q1, Q3) is being used.")

    @data(0, 1, 2)  # TODO: add 3 once https://github.com/Qiskit/qiskit-terra/issues/6406 is fixed
    def test_level(self, level):
        """Test level {level} Ourense backend with a faulty CX(Q1, Q3)"""
        circuit = QuantumCircuit(QuantumRegister(5, "qr"))
        circuit.h(range(5))
        circuit.ccx(0, 1, 2)
        circuit.ccx(2, 3, 4)
        circuit.measure_all()
        result = transpile(
            circuit,
            backend=FakeOurenseFaultyCX13(),
            optimization_level=level,
            initial_layout=range(5),
            seed_transpiler=42,
        )
        self.assertIdleCX13(result)
        self.assertEqualCount(circuit, result)

    @data(0, 1, 2)  # TODO: add 3 once https://github.com/Qiskit/qiskit-terra/issues/6406 is fixed
    def test_layout_level(self, level):
        """Test level {level} with a faulty CX(Q1, Q3) with a working initial layout"""

        #       ┌───┐     ┌───┐                               ░ ┌─┐
        # qr_0: ┤ H ├──■──┤ X ├───────────────────────────────░─┤M├────────────
        #       ├───┤┌─┴─┐└─┬─┘     ┌───┐     ┌───┐           ░ └╥┘┌─┐
        # qr_1: ┤ H ├┤ X ├──■────■──┤ X ├──■──┤ X ├───────────░──╫─┤M├─────────
        #       ├───┤└───┘     ┌─┴─┐└─┬─┘  │  └─┬─┘           ░  ║ └╥┘┌─┐
        # qr_2: ┤ H ├──────────┤ X ├──■────┼────┼─────────────░──╫──╫─┤M├──────
        #       └───┘          └───┘     ┌─┴─┐  │       ┌───┐ ░  ║  ║ └╥┘┌─┐
        # qr_3: ─────────────────────────┤ X ├──■────■──┤ X ├─░──╫──╫──╫─┤M├───
        #                                └───┘     ┌─┴─┐└─┬─┘ ░  ║  ║  ║ └╥┘┌─┐
        # qr_4: ───────────────────────────────────┤ X ├──■───░──╫──╫──╫──╫─┤M├
        #                                          └───┘      ░  ║  ║  ║  ║ └╥┘
        # meas: 5/═══════════════════════════════════════════════╩══╩══╩══╩══╩═
        #                                                        0  1  2  3  4
        circuit = QuantumCircuit(QuantumRegister(5, "qr"))
        circuit.h(range(3))
        circuit.cx(0, 1)
        circuit.cx(1, 0)
        circuit.cx(1, 2)
        circuit.cx(2, 1)
        circuit.cx(1, 3)
        circuit.cx(3, 1)
        circuit.cx(3, 4)
        circuit.cx(4, 3)
        circuit.measure_all()
        result = transpile(
            circuit,
            backend=FakeOurenseFaultyCX13(),
            optimization_level=level,
            initial_layout=[0, 1, 2, 3, 4],
            seed_transpiler=42,
        )

        self.assertIdleCX13(result)
        self.assertEqualCount(circuit, result)


@ddt
class TestFaultyQ1(TestFaultyBackendCase):
    """Test preset passmanagers with FakeOurenseFaultyQ1.
    A 5 qubit backend, with a faulty q1
      0 ↔ (1) ↔ 3 ↔ 4
           ↕
           2
    """

    def assertIdleQ1(self, circuit):
        """Asserts the Q1 in circuit is not used with operations"""
        physical_qubits = QuantumRegister(5, "q")
        nodes = circuit_to_dag(circuit).nodes_on_wire(physical_qubits[1])
        for node in nodes:
            if isinstance(node, DAGOpNode):
                raise AssertionError("Faulty Qubit Q1 not totally idle")

    @data(0, 1, 2, 3)
    def test_level(self, level):
        """Test level {level} Ourense backend with a faulty Q1"""
        circuit = QuantumCircuit(QuantumRegister(2, "qr"))
        circuit.h(range(2))
        circuit.cz(0, 1)
        circuit.measure_all()
        result = transpile(
            circuit, backend=FakeOurenseFaultyQ1(), optimization_level=level, seed_transpiler=42
        )

        self.assertIdleQ1(result)
        self.assertEqualCount(circuit, result)

    @data(0, 1, 2, 3)
    def test_layout_level(self, level):
        """Test level {level} with a faulty Q1 with a working initial layout"""
        circuit = QuantumCircuit(QuantumRegister(2, "qr"))
        circuit.h(range(2))
        circuit.cz(0, 1)
        circuit.measure_all()
        result = transpile(
            circuit,
            backend=FakeOurenseFaultyQ1(),
            optimization_level=level,
            initial_layout=[4, 3],
            seed_transpiler=42,
        )

        self.assertIdleQ1(result)
        self.assertEqualCount(circuit, result)

    @data(0, 1, 2, 3)
    def test_failing_layout_level(self, level):
        """Test level {level} with a faulty Q1 with a failing initial layout. Raises."""
        circuit = QuantumCircuit(QuantumRegister(2, "qr"))
        circuit.h(range(2))
        circuit.cz(0, 1)
        circuit.measure_all()

        message = "The initial_layout parameter refers to faulty or disconnected qubits"

        with self.assertRaises(TranspilerError) as context:
            transpile(
                circuit,
                backend=FakeOurenseFaultyQ1(),
                optimization_level=level,
                initial_layout=[4, 0],
                seed_transpiler=42,
            )

        self.assertEqual(context.exception.message, message)


class TestFaultyQ1Unpickable(TestFaultyBackendCase):
    """See:
    https://github.com/Qiskit/qiskit-terra/pull/4723
    https://github.com/Qiskit/qiskit-terra/pull/4782
    """

    def setUp(self):
        """Creates a FakeBackend that is unpickable"""
        super().setUp()
        backend = FakeOurenseFaultyQ1()
        backend.unpickable_prop = lambda x: x
        self.unpickable_backend = backend

    def test_unpickable_backend(self):
        """Test Ourense unpickable backend with a faulty Q1 in parallel"""
        circuit = QuantumCircuit(QuantumRegister(2, "qr"))
        circuit.h(range(2))
        circuit.cz(0, 1)
        circuit.measure_all()
        result = transpile(
            [circuit, circuit],
            backend=self.unpickable_backend,
            optimization_level=1,
            seed_transpiler=42,
        )
        self.assertEqualCount(circuit, result[0])
        self.assertEqualCount(circuit, result[1])



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2018.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Tests for the wrapper functionality."""

import io
import unittest
from qiskit import QuantumRegister, ClassicalRegister, QuantumCircuit
from qiskit import BasicAer
from qiskit import execute
from qiskit.tools.monitor import job_monitor
from qiskit.test import QiskitTestCase


class TestJobMonitor(QiskitTestCase):
    """Tools test case."""

    def test_job_monitor(self):
        """Test job_monitor"""
        qreg = QuantumRegister(2)
        creg = ClassicalRegister(2)
        qc = QuantumCircuit(qreg, creg)
        qc.h(qreg[0])
        qc.cx(qreg[0], qreg[1])
        qc.measure(qreg, creg)
        backend = BasicAer.get_backend("qasm_simulator")
        job_sim = execute([qc] * 10, backend)
        output = io.StringIO()
        job_monitor(job_sim, output=output)
        self.assertEqual(job_sim.status().name, "DONE")


if __name__ == "__main__":
    unittest.main(verbosity=2)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2018.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
=========================
Qobj (:mod:`qiskit.qobj`)
=========================

.. currentmodule:: qiskit.qobj

Base
====

.. autosummary::
   :toctree: ../stubs/

   Qobj
   QobjExperimentHeader
   QobjHeader

Qasm
====

.. autosummary::
   :toctree: ../stubs/

   QasmQobj
   QasmQobjInstruction
   QasmQobjExperimentConfig
   QasmQobjExperiment
   QasmQobjConfig
   QasmExperimentCalibrations
   GateCalibration

Pulse
=====

.. autosummary::
   :toctree: ../stubs/

   PulseQobj
   PulseQobjInstruction
   PulseQobjExperimentConfig
   PulseQobjExperiment
   PulseQobjConfig
   QobjMeasurementOption
   PulseLibraryItem
"""

import warnings

from qiskit.qobj.common import QobjExperimentHeader
from qiskit.qobj.common import QobjHeader

from qiskit.qobj.pulse_qobj import PulseQobj
from qiskit.qobj.pulse_qobj import PulseQobjInstruction
from qiskit.qobj.pulse_qobj import PulseQobjExperimentConfig
from qiskit.qobj.pulse_qobj import PulseQobjExperiment
from qiskit.qobj.pulse_qobj import PulseQobjConfig
from qiskit.qobj.pulse_qobj import QobjMeasurementOption
from qiskit.qobj.pulse_qobj import PulseLibraryItem

from qiskit.qobj.qasm_qobj import GateCalibration
from qiskit.qobj.qasm_qobj import QasmExperimentCalibrations
from qiskit.qobj.qasm_qobj import QasmQobj
from qiskit.qobj.qasm_qobj import QasmQobjInstruction
from qiskit.qobj.qasm_qobj import QasmQobjExperiment
from qiskit.qobj.qasm_qobj import QasmQobjConfig
from qiskit.qobj.qasm_qobj import QasmQobjExperimentConfig


class Qobj(QasmQobj):
    """A backwards compat alias for QasmQobj."""

    def __init__(self, qobj_id=None, config=None, experiments=None, header=None):
        """Initialize a Qobj object."""
        warnings.warn(
            "qiskit.qobj.Qobj is deprecated use either QasmQobj or "
            "PulseQobj depending on your application instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(qobj_id=qobj_id, config=config, experiments=experiments, header=header)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

""" OperatorBase Class """

import itertools
from abc import ABC, abstractmethod
from copy import deepcopy
from typing import Dict, List, Optional, Set, Tuple, Union, cast

import numpy as np
from scipy.sparse import csr_matrix, spmatrix

from qiskit.circuit import ParameterExpression, ParameterVector
from qiskit.opflow.exceptions import OpflowError
from qiskit.opflow.mixins import StarAlgebraMixin, TensorMixin
from qiskit.quantum_info import Statevector
from qiskit.utils import algorithm_globals


class OperatorBase(StarAlgebraMixin, TensorMixin, ABC):
    """A base class for all Operators: PrimitiveOps, StateFns, ListOps, etc. Operators are
    defined as functions which take one complex binary function to another. These complex binary
    functions are represented by StateFns, which are themselves a special class of Operators
    taking only the ``Zero`` StateFn to the complex binary function they represent.

    Operators can be used to construct complicated functions and computation, and serve as the
    building blocks for algorithms.

    """

    # Indentation used in string representation of list operators
    # Can be changed to use another indentation than two whitespaces
    INDENTATION = "  "

    _count = itertools.count()

    def __init__(self) -> None:
        self._instance_id = next(self._count)

    @property
    @abstractmethod
    def settings(self) -> Dict:
        """Return settings of this object in a dictionary.

        You can, for example, use this ``settings`` dictionary to serialize the
        object in JSON format, if the JSON encoder you use supports all types in
        the dictionary.

        Returns:
            Object settings in a dictionary.
        """
        raise NotImplementedError

    @property
    def instance_id(self) -> int:
        """Return the unique instance id."""
        return self._instance_id

    @property
    @abstractmethod
    def num_qubits(self) -> int:
        r"""The number of qubits over which the Operator is defined. If
        ``op.num_qubits == 5``, then ``op.eval('1' * 5)`` will be valid, but
        ``op.eval('11')`` will not.

        Returns:
            The number of qubits accepted by the Operator's underlying function.
        """
        raise NotImplementedError

    @abstractmethod
    def primitive_strings(self) -> Set[str]:
        r"""Return a set of strings describing the primitives contained in the Operator. For
        example, ``{'QuantumCircuit', 'Pauli'}``. For hierarchical Operators, such as ``ListOps``,
        this can help illuminate the primitives represented in the various recursive levels,
        and therefore which conversions can be applied.

        Returns:
            A set of strings describing the primitives contained within the Operator.
        """
        raise NotImplementedError

    @abstractmethod
    def eval(
        self,
        front: Optional[
            Union[str, Dict[str, complex], np.ndarray, "OperatorBase", Statevector]
        ] = None,
    ) -> Union["OperatorBase", complex]:
        r"""
        Evaluate the Operator's underlying function, either on a binary string or another Operator.
        A square binary Operator can be defined as a function taking a binary function to another
        binary function. This method returns the value of that function for a given StateFn or
        binary string. For example, ``op.eval('0110').eval('1110')`` can be seen as querying the
        Operator's matrix representation by row 6 and column 14, and will return the complex
        value at those "indices." Similarly for a StateFn, ``op.eval('1011')`` will return the
        complex value at row 11 of the vector representation of the StateFn, as all StateFns are
        defined to be evaluated from Zero implicitly (i.e. it is as if ``.eval('0000')`` is already
        called implicitly to always "indexing" from column 0).

        If ``front`` is None, the matrix-representation of the operator is returned.

        Args:
            front: The bitstring, dict of bitstrings (with values being coefficients), or
                StateFn to evaluated by the Operator's underlying function, or None.

        Returns:
            The output of the Operator's evaluation function. If self is a ``StateFn``, the result
            is a float or complex. If self is an Operator (``PrimitiveOp, ComposedOp, SummedOp,
            EvolvedOp,`` etc.), the result is a StateFn.
            If ``front`` is None, the matrix-representation of the operator is returned, which
            is a ``MatrixOp`` for the operators and a ``VectorStateFn`` for state-functions.
            If either self or front contain proper
            ``ListOps`` (not ListOp subclasses), the result is an n-dimensional list of complex
            or StateFn results, resulting from the recursive evaluation by each OperatorBase
            in the ListOps.

        """
        raise NotImplementedError

    @abstractmethod
    def reduce(self):
        r"""Try collapsing the Operator structure, usually after some type of conversion,
        e.g. trying to add Operators in a SummedOp or delete needless IGates in a CircuitOp.
        If no reduction is available, just returns self.

        Returns:
            The reduced ``OperatorBase``.
        """
        raise NotImplementedError

    @abstractmethod
    def to_matrix(self, massive: bool = False) -> np.ndarray:
        r"""Return NumPy representation of the Operator. Represents the evaluation of
        the Operator's underlying function on every combination of basis binary strings.
        Warn if more than 16 qubits to force having to set ``massive=True`` if such a
        large vector is desired.

        Returns:
              The NumPy ``ndarray`` equivalent to this Operator.
        """
        raise NotImplementedError

    @abstractmethod
    def to_matrix_op(self, massive: bool = False) -> "OperatorBase":
        """Returns a ``MatrixOp`` equivalent to this Operator."""
        raise NotImplementedError

    @abstractmethod
    def to_circuit_op(self) -> "OperatorBase":
        """Returns a ``CircuitOp`` equivalent to this Operator."""
        raise NotImplementedError

    def to_spmatrix(self) -> spmatrix:
        r"""Return SciPy sparse matrix representation of the Operator. Represents the evaluation of
        the Operator's underlying function on every combination of basis binary strings.

        Returns:
              The SciPy ``spmatrix`` equivalent to this Operator.
        """
        return csr_matrix(self.to_matrix())

    def is_hermitian(self) -> bool:
        """Return True if the operator is hermitian.

        Returns: Boolean value
        """
        return (self.to_spmatrix() != self.to_spmatrix().getH()).nnz == 0

    @staticmethod
    def _indent(lines: str, indentation: str = INDENTATION) -> str:
        """Indented representation to allow pretty representation of nested operators."""
        indented_str = indentation + lines.replace("\n", f"\n{indentation}")
        if indented_str.endswith(f"\n{indentation}"):
            indented_str = indented_str[: -len(indentation)]
        return indented_str

    # Addition / Subtraction

    @abstractmethod
    def add(self, other: "OperatorBase") -> "OperatorBase":
        r"""Return Operator addition of self and other, overloaded by ``+``.

        Args:
            other: An ``OperatorBase`` with the same number of qubits as self, and in the same
                'Operator', 'State function', or 'Measurement' category as self (i.e. the same type
                of underlying function).

        Returns:
            An ``OperatorBase`` equivalent to the sum of self and other.
        """
        raise NotImplementedError

    # Negation

    def neg(self) -> "OperatorBase":
        r"""Return the Operator's negation, effectively just multiplying by -1.0,
        overloaded by ``-``.

        Returns:
            An ``OperatorBase`` equivalent to the negation of self.
        """
        return self.mul(-1.0)

    # Adjoint

    @abstractmethod
    def adjoint(self) -> "OperatorBase":
        r"""Return a new Operator equal to the Operator's adjoint (conjugate transpose),
        overloaded by ``~``. For StateFns, this also turns the StateFn into a measurement.

        Returns:
            An ``OperatorBase`` equivalent to the adjoint of self.
        """
        raise NotImplementedError

    # Equality

    def __eq__(self, other: object) -> bool:
        r"""Overload ``==`` operation to evaluate equality between Operators.

        Args:
            other: The ``OperatorBase`` to compare to self.

        Returns:
            A bool equal to the equality of self and other.
        """
        if not isinstance(other, OperatorBase):
            return NotImplemented
        return self.equals(cast(OperatorBase, other))

    @abstractmethod
    def equals(self, other: "OperatorBase") -> bool:
        r"""
        Evaluate Equality between Operators, overloaded by ``==``. Only returns True if self and
        other are of the same representation (e.g. a DictStateFn and CircuitStateFn will never be
        equal, even if their vector representations are equal), their underlying primitives are
        equal (this means for ListOps, OperatorStateFns, or EvolvedOps the equality is evaluated
        recursively downwards), and their coefficients are equal.

        Args:
            other: The ``OperatorBase`` to compare to self.

        Returns:
            A bool equal to the equality of self and other.

        """
        raise NotImplementedError

    # Scalar Multiplication

    # pylint: disable=arguments-differ
    @abstractmethod
    def mul(self, scalar: Union[complex, ParameterExpression]) -> "OperatorBase":
        r"""
        Returns the scalar multiplication of the Operator, overloaded by ``*``, including
        support for Terra's ``Parameters``, which can be bound to values later (via
        ``bind_parameters``).

        Args:
            scalar: The real or complex scalar by which to multiply the Operator,
                or the ``ParameterExpression`` to serve as a placeholder for a scalar factor.

        Returns:
            An ``OperatorBase`` equivalent to product of self and scalar.
        """
        raise NotImplementedError

    @abstractmethod
    def tensor(self, other: "OperatorBase") -> "OperatorBase":
        r"""Return tensor product between self and other, overloaded by ``^``.
        Note: You must be conscious of Qiskit's big-endian bit printing convention.
        Meaning, X.tensor(Y) produces an X on qubit 0 and an Y on qubit 1, or X⨂Y,
        but would produce a QuantumCircuit which looks like

            -[Y]-
            -[X]-

        Because Terra prints circuits and results with qubit 0 at the end of the string
        or circuit.

        Args:
            other: The ``OperatorBase`` to tensor product with self.

        Returns:
            An ``OperatorBase`` equivalent to the tensor product of self and other.
        """
        raise NotImplementedError

    @abstractmethod
    def tensorpower(self, other: int) -> Union["OperatorBase", int]:
        r"""Return tensor product with self multiple times, overloaded by ``^``.

        Args:
            other: The int number of times to tensor product self with itself via ``tensorpower``.

        Returns:
            An ``OperatorBase`` equivalent to the tensorpower of self by other.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def parameters(self):
        r"""Return a set of Parameter objects contained in the Operator."""
        raise NotImplementedError

    # Utility functions for parameter binding

    @abstractmethod
    def assign_parameters(
        self,
        param_dict: Dict[
            ParameterExpression,
            Union[complex, ParameterExpression, List[Union[complex, ParameterExpression]]],
        ],
    ) -> "OperatorBase":
        """Binds scalar values to any Terra ``Parameters`` in the coefficients or primitives of
        the Operator, or substitutes one ``Parameter`` for another. This method differs from
        Terra's ``assign_parameters`` in that it also supports lists of values to assign for a
        give ``Parameter``, in which case self will be copied for each parameterization in the
        binding list(s), and all the copies will be returned in an ``OpList``. If lists of
        parameterizations are used, every ``Parameter`` in the param_dict must have the same
        length list of parameterizations.

        Args:
            param_dict: The dictionary of ``Parameters`` to replace, and values or lists of
                values by which to replace them.

        Returns:
            The ``OperatorBase`` with the ``Parameters`` in self replaced by the
            values or ``Parameters`` in param_dict. If param_dict contains parameterization lists,
            this ``OperatorBase`` is an ``OpList``.
        """
        raise NotImplementedError

    @abstractmethod
    def _expand_dim(self, num_qubits: int) -> "OperatorBase":
        """Expands the operator with identity operator of dimension 2**num_qubits.

        Returns:
            Operator corresponding to self.tensor(identity_operator), where dimension of identity
            operator is 2 ** num_qubits.
        """
        raise NotImplementedError

    @abstractmethod
    def permute(self, permutation: List[int]) -> "OperatorBase":
        """Permutes the qubits of the operator.

        Args:
            permutation: A list defining where each qubit should be permuted. The qubit at index
                j should be permuted to position permutation[j].

        Returns:
            A new OperatorBase containing the permuted operator.

        Raises:
            OpflowError: if indices do not define a new index for each qubit.
        """
        raise NotImplementedError

    def bind_parameters(
        self,
        param_dict: Dict[
            ParameterExpression,
            Union[complex, ParameterExpression, List[Union[complex, ParameterExpression]]],
        ],
    ) -> "OperatorBase":
        r"""
        Same as assign_parameters, but maintained for consistency with QuantumCircuit in
        Terra (which has both assign_parameters and bind_parameters).
        """
        return self.assign_parameters(param_dict)

    # Mostly copied from terra, but with list unrolling added:
    @staticmethod
    def _unroll_param_dict(
        value_dict: Dict[Union[ParameterExpression, ParameterVector], Union[complex, List[complex]]]
    ) -> Union[Dict[ParameterExpression, complex], List[Dict[ParameterExpression, complex]]]:
        """Unrolls the ParameterVectors in a param_dict into separate Parameters, and unrolls
        parameterization value lists into separate param_dicts without list nesting."""
        unrolled_value_dict = {}
        for (param, value) in value_dict.items():
            if isinstance(param, ParameterExpression):
                unrolled_value_dict[param] = value
            if isinstance(param, ParameterVector) and isinstance(value, (list, np.ndarray)):
                if not len(param) == len(value):
                    raise ValueError(
                        "ParameterVector {} has length {}, which differs from value list {} of "
                        "len {}".format(param, len(param), value, len(value))
                    )
                unrolled_value_dict.update(zip(param, value))
        if isinstance(list(unrolled_value_dict.values())[0], list):
            # check that all are same length
            unrolled_value_dict_list = []
            try:
                for i in range(len(list(unrolled_value_dict.values())[0])):  # type: ignore
                    unrolled_value_dict_list.append(
                        OperatorBase._get_param_dict_for_index(
                            unrolled_value_dict, i  # type: ignore
                        )
                    )
                return unrolled_value_dict_list
            except IndexError as ex:
                raise OpflowError("Parameter binding lists must all be the same length.") from ex
        return unrolled_value_dict  # type: ignore

    @staticmethod
    def _get_param_dict_for_index(unrolled_dict: Dict[ParameterExpression, List[complex]], i: int):
        """Gets a single non-list-nested param_dict for a given list index from a nested one."""
        return {k: v[i] for (k, v) in unrolled_dict.items()}

    def _expand_shorter_operator_and_permute(
        self, other: "OperatorBase", permutation: Optional[List[int]] = None
    ) -> Tuple["OperatorBase", "OperatorBase"]:
        if permutation is not None:
            other = other.permute(permutation)
        new_self = self
        if not self.num_qubits == other.num_qubits:
            # pylint: disable=cyclic-import
            from .operator_globals import Zero

            if other == Zero:
                # Zero is special - we'll expand it to the correct qubit number.
                other = Zero.__class__("0" * self.num_qubits)
            elif other.num_qubits < self.num_qubits:
                other = other._expand_dim(self.num_qubits - other.num_qubits)
            elif other.num_qubits > self.num_qubits:
                new_self = self._expand_dim(other.num_qubits - self.num_qubits)
        return new_self, other

    def copy(self) -> "OperatorBase":
        """Return a deep copy of the Operator."""
        return deepcopy(self)

    # Composition

    # pylint: disable=arguments-differ
    @abstractmethod
    def compose(
        self, other: "OperatorBase", permutation: Optional[List[int]] = None, front: bool = False
    ) -> "OperatorBase":
        r"""Return Operator Composition between self and other (linear algebra-style:
        A@B(x) = A(B(x))), overloaded by ``@``.

        Note: You must be conscious of Quantum Circuit vs. Linear Algebra ordering
        conventions. Meaning, X.compose(Y)
        produces an X∘Y on qubit 0, but would produce a QuantumCircuit which looks like

            -[Y]-[X]-

        Because Terra prints circuits with the initial state at the left side of the circuit.

        Args:
            other: The ``OperatorBase`` with which to compose self.
            permutation: ``List[int]`` which defines permutation on other operator.
            front: If front==True, return ``other.compose(self)``.

        Returns:
            An ``OperatorBase`` equivalent to the function composition of self and other.
        """
        raise NotImplementedError

    @staticmethod
    def _check_massive(method: str, matrix: bool, num_qubits: int, massive: bool) -> None:
        """
        Checks if matrix or vector generated will be too large.

        Args:
            method: Name of the calling method
            matrix: True if object is matrix, otherwise vector
            num_qubits: number of qubits
            massive: True if it is ok to proceed with large matrix

        Raises:
            ValueError: Massive is False and number of qubits is greater than 16
        """
        if num_qubits > 16 and not massive and not algorithm_globals.massive:
            dim = 2**num_qubits
            if matrix:
                obj_type = "matrix"
                dimensions = f"{dim}x{dim}"
            else:
                obj_type = "vector"
                dimensions = f"{dim}"
            raise ValueError(
                f"'{method}' will return an exponentially large {obj_type}, "
                f"in this case '{dimensions}' elements. "
                "Set algorithm_globals.massive=True or the method argument massive=True "
                "if you want to proceed."
            )

    # Printing

    @abstractmethod
    def __str__(self) -> str:
        raise NotImplementedError



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Operator Evolutions (:mod:`qiskit.opflow.evolutions`)
=====================================================

.. currentmodule:: qiskit.opflow.evolutions

Evolutions are converters which traverse an Operator tree, replacing any :class:`EvolvedOp` `e`
with a Schrodinger equation-style evolution :class:`~qiskit.opflow.primitive_ops.CircuitOp`
equalling or approximating the matrix exponential of -i * the Operator contained inside
(`e.primitive`). The Evolutions are essentially implementations of Hamiltonian Simulation
algorithms, including various methods for Trotterization.

The :class:`EvolvedOp` is simply a placeholder signifying that the Operator inside it should be
converted to its exponential by the Evolution converter. All Operators
(not :mod:`~qiskit.opflow.state_fns`) have
``.exp_i()`` methods which either return the exponential of the Operator directly,
or an :class:`EvolvedOp` containing the Operator.

Note:
    Evolutions work with parameterized Operator coefficients, so
    ``my_expectation.convert((t * H).exp_i())``, where t is a scalar or Terra Parameter and H
    is an Operator, will produce a :class:`~qiskit.opflow.primitive_ops.CircuitOp`
    equivalent to e^iHt.

Evolution Base Class
--------------------

The EvolutionBase class gives an interface for algorithms to ask for Evolutions as
execution settings. For example, if an algorithm contains an Operator evolution step within it,
such as :class:`~qiskit.algorithms.QAOA`, the algorithm can give the opportunity for the user
to pass an EvolutionBase of their choice to be used in that evolution step.

.. autosummary::
   :toctree: ../stubs/
   :template: autosummary/class_no_inherited_members.rst

   EvolutionBase

Evolutions
----------

.. autosummary::
   :toctree: ../stubs/
   :template: autosummary/class_no_inherited_members.rst

   EvolutionFactory
   EvolvedOp
   MatrixEvolution
   PauliTrotterEvolution

Trotterizations
---------------

.. autosummary::
   :toctree: ../stubs/
   :template: autosummary/class_no_inherited_members.rst

   TrotterizationBase
   TrotterizationFactory
   Trotter
   Suzuki
   QDrift
"""

from .evolution_base import EvolutionBase
from .evolution_factory import EvolutionFactory
from .evolved_op import EvolvedOp
from .pauli_trotter_evolution import PauliTrotterEvolution
from .matrix_evolution import MatrixEvolution
from .trotterizations import TrotterizationBase, TrotterizationFactory, Trotter, Suzuki, QDrift

# TODO co-diagonalization of Abelian groups in PauliTrotterEvolution
# TODO quantum signal processing/qubitization
# TODO evolve by density matrix (need to add iexp to operator_state_fn)
# TODO linear combination evolution

__all__ = [
    "EvolutionBase",
    "EvolutionFactory",
    "EvolvedOp",
    "PauliTrotterEvolution",
    "MatrixEvolution",
    "TrotterizationBase",
    "TrotterizationFactory",
    "Trotter",
    "Suzuki",
    "QDrift",
]



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Methods for generating random quantum information objects."""

# pylint: disable=unused-import
from qiskit.quantum_info.operators.random import (
    random_clifford,
    random_cnotdihedral,
    random_hermitian,
    random_pauli,
    random_pauli_list,
    random_pauli_table,
    random_quantum_channel,
    random_stabilizer_table,
    random_unitary,
)
from qiskit.quantum_info.states.random import random_density_matrix, random_statevector



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Functions for dealing with classical conditions."""

from typing import Tuple, Union

from qiskit.circuit.classicalregister import ClassicalRegister, Clbit
from qiskit.circuit.exceptions import CircuitError


def validate_condition(
    condition: Tuple[Union[ClassicalRegister, Clbit], int]
) -> Tuple[Union[ClassicalRegister, Clbit], int]:
    """Validate that a condition is in a valid format and return it, but raise if it is invalid.

    Args:
        condition: the condition to be tested for validity.

    Raises:
        CircuitError: if the condition is not in a valid format.

    Returns:
        The same condition as passed, if it was valid.
    """
    try:
        bits, value = condition
        if isinstance(bits, (ClassicalRegister, Clbit)) and isinstance(value, int):
            return (bits, value)
    except (TypeError, ValueError):
        pass
    raise CircuitError(
        "A classical condition should be a 2-tuple of `(ClassicalRegister | Clbit, int)`,"
        f" but received '{condition!r}'."
    )


def condition_bits(condition: Tuple[Union[ClassicalRegister, Clbit], int]) -> Tuple[Clbit, ...]:
    """Return the classical resources used by ``condition`` as a tuple of :obj:`.Clbit`.

    This is useful when the exact set of bits is required, rather than the logical grouping of
    :obj:`.ClassicalRegister`, such as when determining circuit blocking.

    Args:
        condition: the valid condition to extract the bits from.

    Returns:
        a tuple of all classical bits used in the condition.
    """
    return (condition[0],) if isinstance(condition[0], Clbit) else tuple(condition[0])


def condition_registers(
    condition: Tuple[Union[ClassicalRegister, Clbit], int]
) -> Tuple[ClassicalRegister, ...]:
    """Return any classical registers used by ``condition`` as a tuple of :obj:`.ClassicalRegister`.

    This is useful as a quick method for extracting the registers from a condition, if any exist.
    The output might be empty if the condition is on a single bit.

    Args:
        condition: the valid condition to extract any registers from.

    Returns:
        a tuple of all classical registers used in the condition.
    """
    return (condition[0],) if isinstance(condition[0], ClassicalRegister) else ()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Padding pass to insert Delay to the empty slots."""

from qiskit.circuit import Qubit
from qiskit.circuit.delay import Delay
from qiskit.dagcircuit import DAGCircuit, DAGNode, DAGOutNode

from .base_padding import BasePadding


class PadDelay(BasePadding):
    """Padding idle time with Delay instructions.

    Consecutive delays will be merged in the output of this pass.

    .. code-block:: python

        durations = InstructionDurations([("x", None, 160), ("cx", None, 800)])

        qc = QuantumCircuit(2)
        qc.delay(100, 0)
        qc.x(1)
        qc.cx(0, 1)

    The ASAP-scheduled circuit output may become

    .. parsed-literal::

             ┌────────────────┐
        q_0: ┤ Delay(160[dt]) ├──■──
             └─────┬───┬──────┘┌─┴─┐
        q_1: ──────┤ X ├───────┤ X ├
                   └───┘       └───┘

    Note that the additional idle time of 60dt on the ``q_0`` wire coming from the duration difference
    between ``Delay`` of 100dt (``q_0``) and ``XGate`` of 160 dt (``q_1``) is absorbed in
    the delay instruction on the ``q_0`` wire, i.e. in total 160 dt.

    See :class:`BasePadding` pass for details.
    """

    def __init__(self, fill_very_end: bool = True):
        """Create new padding delay pass.

        Args:
            fill_very_end: Set ``True`` to fill the end of circuit with delay.
        """
        super().__init__()
        self.fill_very_end = fill_very_end

    def _pad(
        self,
        dag: DAGCircuit,
        qubit: Qubit,
        t_start: int,
        t_end: int,
        next_node: DAGNode,
        prev_node: DAGNode,
    ):
        if not self.fill_very_end and isinstance(next_node, DAGOutNode):
            return

        time_interval = t_end - t_start
        self._apply_scheduled_op(dag, t_start, Delay(time_interval, dag.unit), qubit)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""The module for Quantum the Fisher Information."""
import copy
from typing import List, Union

import numpy as np
from qiskit.circuit import ParameterVector, ParameterExpression
from qiskit.circuit.library import RZGate, RXGate, RYGate
from qiskit.converters import dag_to_circuit, circuit_to_dag
from ...list_ops.list_op import ListOp
from ...primitive_ops.circuit_op import CircuitOp
from ...expectations.pauli_expectation import PauliExpectation
from ...operator_globals import I, Z, Y, X, Zero
from ...state_fns.state_fn import StateFn
from ...state_fns.circuit_state_fn import CircuitStateFn


from .circuit_qfi import CircuitQFI
from ..derivative_base import _coeff_derivative


class OverlapDiag(CircuitQFI):
    r"""Compute the diagonal of the QFI given a pure, parameterized quantum state.

    See also :class:`~qiskit.opflow.QFI`.
    """

    def convert(
        self,
        operator: Union[CircuitOp, CircuitStateFn],
        params: Union[ParameterExpression, ParameterVector, List[ParameterExpression]],
    ) -> ListOp:
        r"""
        Args:
            operator: The operator corresponding to the quantum state :math:`|\psi(\omega)\rangle`
                for which we compute the QFI.
            params: The parameters :math:`\omega` with respect to which we are computing the QFI.

        Returns:
            A ``ListOp[ListOp]`` where the operator at position ``[k][l]`` corresponds to the matrix
            element :math:`k, l` of the QFI.

        Raises:
            NotImplementedError: If ``operator`` is neither ``CircuitOp`` nor ``CircuitStateFn``.

        """

        if not isinstance(operator, CircuitStateFn):
            raise NotImplementedError("operator must be a CircuitStateFn")

        return self._diagonal_approx(operator=operator, params=params)

    # TODO, for some reason diagonal_approx doesn't use the same get_parameter_expression method.
    # This should be fixed.
    def _diagonal_approx(
        self,
        operator: Union[CircuitOp, CircuitStateFn],
        params: Union[ParameterExpression, ParameterVector, List],
    ) -> ListOp:
        """
        Args:
            operator: The operator corresponding to the quantum state |ψ(ω)〉for which we compute
                the QFI
            params: The parameters we are computing the QFI wrt: ω

        Returns:
            ListOp where the operator at position k corresponds to QFI_k,k

        Raises:
            NotImplementedError: If a circuit is found such that one parameter controls multiple
                                 gates, or one gate contains multiple parameters.
            TypeError: If a circuit is found that includes more than one parameter as they are
                       currently not supported for the overlap diagonal QFI method.

        """

        if not isinstance(operator, (CircuitOp, CircuitStateFn)):
            raise NotImplementedError("operator must be a CircuitOp or CircuitStateFn")

        # If a single parameter is given wrap it into a list.
        if isinstance(params, ParameterExpression):
            params = [params]

        circuit = operator.primitive

        # Partition the circuit into layers, and build the circuits to prepare $\psi_i$
        layers = _partition_circuit(circuit)
        if layers[-1].num_parameters == 0:
            layers.pop(-1)

        psis = [CircuitOp(layer) for layer in layers]
        for i, psi in enumerate(psis):
            if i == 0:
                continue
            psis[i] = psi @ psis[i - 1]

        # TODO: make this work for other types of rotations
        # NOTE: This assumes that each parameter only affects one rotation.
        # we need to think more about what happens if multiple rotations
        # are controlled with a single parameter.
        generators = _get_generators(params, circuit)

        diag = []
        for param in params:
            if len(circuit._parameter_table[param]) > 1:
                raise NotImplementedError(
                    "OverlapDiag does not yet support multiple "
                    "gates parameterized by a single parameter. For such "
                    "circuits use LinCombFull"
                )

            gate = next(iter(circuit._parameter_table[param]))[0]

            if len(gate.params) != 1:
                raise TypeError(
                    "OverlapDiag cannot yet support gates with more than one parameter."
                )

            param_value = gate.params[0]
            generator = generators[param]
            meas_op = ~StateFn(generator)

            # get appropriate psi_i
            psi = [(psi) for psi in psis if param in psi.primitive.parameters][0]

            op = meas_op @ psi @ Zero
            if type(param_value) == ParameterExpression:  # pylint: disable=unidiomatic-typecheck
                expr_grad = _coeff_derivative(param_value, param)
                op *= expr_grad
            rotated_op = PauliExpectation().convert(op)
            diag.append(rotated_op)

        grad_op = ListOp(diag, combo_fn=lambda x: np.diag(np.real([1 - y**2 for y in x])))
        return grad_op


def _partition_circuit(circuit):
    dag = circuit_to_dag(circuit)
    dag_layers = [i["graph"] for i in dag.serial_layers()]
    num_qubits = circuit.num_qubits
    layers = list(
        zip(dag_layers, [{x: False for x in range(0, num_qubits)} for layer in dag_layers])
    )

    # initialize the ledger
    # The ledger tracks which qubits in each layer are available to have
    # gates from subsequent layers shifted backward.
    # The idea being that all parameterized gates should have
    # no descendants within their layer
    bit_indices = {bit: index for index, bit in enumerate(circuit.qubits)}
    for i, (layer, ledger) in enumerate(layers):
        op_node = layer.op_nodes()[0]
        is_param = op_node.op.is_parameterized()
        qargs = op_node.qargs
        indices = [bit_indices[qarg] for qarg in qargs]
        if is_param:
            for index in indices:
                ledger[index] = True

    def apply_node_op(node, dag, back=True):
        op = copy.copy(node.op)
        qargs = copy.copy(node.qargs)
        cargs = copy.copy(node.cargs)
        if back:
            dag.apply_operation_back(op, qargs, cargs)
        else:
            dag.apply_operation_front(op, qargs, cargs)

    converged = False

    for _ in range(dag.depth() + 1):
        if converged:
            break

        converged = True

        for i, (layer, ledger) in enumerate(layers):
            if i == len(layers) - 1:
                continue

            (next_layer, next_ledger) = layers[i + 1]
            for next_node in next_layer.op_nodes():
                is_param = next_node.op.is_parameterized()
                qargs = next_node.qargs
                indices = [bit_indices[qarg] for qarg in qargs]

                # If the next_node can be moved back a layer without
                # without becoming the descendant of a parameterized gate,
                # then do it.
                if not any(ledger[x] for x in indices):

                    apply_node_op(next_node, layer)
                    next_layer.remove_op_node(next_node)

                    if is_param:
                        for index in indices:
                            ledger[index] = True
                            next_ledger[index] = False

                    converged = False

            # clean up empty layers left behind.
            if len(next_layer.op_nodes()) == 0:
                layers.pop(i + 1)

    partitioned_circs = [dag_to_circuit(layer[0]) for layer in layers]
    return partitioned_circs


def _get_generators(params, circuit):
    dag = circuit_to_dag(circuit)
    layers = list(dag.serial_layers())

    generators = {}
    num_qubits = dag.num_qubits()
    bit_indices = {bit: index for index, bit in enumerate(circuit.qubits)}

    for layer in layers:
        instr = layer["graph"].op_nodes()[0].op
        # if no gate is parameterized, skip
        if not any(isinstance(param, ParameterExpression) for param in instr.params):
            continue

        if len(instr.params) != 1:
            raise NotImplementedError(
                "The QFI diagonal approximation currently only supports "
                "gates with a single free parameter."
            )
        param_value = instr.params[0]

        for param in params:
            if param in param_value.parameters:

                if isinstance(instr, RYGate):
                    generator = Y
                elif isinstance(instr, RZGate):
                    generator = Z
                elif isinstance(instr, RXGate):
                    generator = X
                else:
                    raise NotImplementedError(f"Generator for gate {instr.name} not implemented.")

                # get all qubit indices in this layer where the param parameterizes
                # an operation.
                indices = [[bit_indices[q] for q in qreg] for qreg in layer["partition"]]
                indices = [item for sublist in indices for item in sublist]

                if len(indices) > 1:
                    raise NotImplementedError
                index = indices[0]
                generator = (I ^ (index)) ^ generator ^ (I ^ (num_qubits - index - 1))
                generators[param] = generator

    return generators



# This code is part of Qiskit.
#
# (C) Copyright IBM 2019, 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""The projected Variational Quantum Dynamics Algorithm."""
from __future__ import annotations

import logging
from typing import Callable

import numpy as np

from qiskit.circuit import QuantumCircuit, ParameterVector, Parameter
from qiskit.circuit.library import PauliEvolutionGate
from qiskit.opflow import PauliSumOp
from qiskit.primitives import BaseEstimator
from qiskit.quantum_info.operators.base_operator import BaseOperator
from qiskit.synthesis import EvolutionSynthesis, LieTrotter
from ...exceptions import AlgorithmError, QiskitError
from .pvqd_result import PVQDResult
from .utils import _get_observable_evaluator, _is_gradient_supported
from ..time_evolution_problem import TimeEvolutionProblem
from ..time_evolution_result import TimeEvolutionResult
from ..real_time_evolver import RealTimeEvolver
from ...state_fidelities.base_state_fidelity import BaseStateFidelity
from ...optimizers import Optimizer, Minimizer

logger = logging.getLogger(__name__)


class PVQD(RealTimeEvolver):
    """The projected Variational Quantum Dynamics (p-VQD) Algorithm.

    In each timestep, this algorithm computes the next state with a Trotter formula
    (specified by the ``evolution`` argument) and projects the timestep onto a variational form
    (``ansatz``). The projection is determined by maximizing the fidelity of the Trotter-evolved
    state and the ansatz, using a classical optimization routine. See Ref. [1] for details.

    The following attributes can be set via the initializer but can also be read and
    updated once the PVQD object has been constructed.

    Attributes:

        ansatz (QuantumCircuit): The parameterized circuit representing the time-evolved state.
        initial_parameters (np.ndarray): The parameters of the ansatz at time 0.
        optimizer (Optional[Union[Optimizer, Minimizer]]): The classical optimization routine
            used to maximize the fidelity of the Trotter step and ansatz.
        num_timesteps (Optional[int]): The number of timesteps to take. If None, it is automatically
            selected to achieve a timestep of approximately 0.01.
        evolution (Optional[EvolutionSynthesis]): The method to perform the Trotter step.
            Defaults to first-order Lie-Trotter evolution.
        use_parameter_shift (bool): If True, use the parameter shift rule for loss function
            gradients (if the ansatz supports).
        initial_guess (Optional[np.ndarray]): The starting point for the first classical optimization
            run, at time 0. Defaults to random values in :math:`[-0.01, 0.01]`.

    Example:

        This snippet computes the real time evolution of a quantum Ising model on two
        neighboring sites and keeps track of the magnetization.

        .. code-block:: python

            import numpy as np

            from qiskit.algorithms.state_fidelities import ComputeUncompute
            from qiskit.algorithms.time_evolvers.pvqd import PVQD
            from qiskit.primitives import Estimator
            from qiskit import BasicAer
            from qiskit.circuit.library import EfficientSU2
            from qiskit.quantum_info import Pauli, SparsePauliOp
            from qiskit.algorithms.optimizers import L_BFGS_B

            sampler = Sampler()
            fidelity = ComputeUncompute(sampler)
            estimator = Estimator()
            hamiltonian = 0.1 * SparsePauliOp([Pauli("ZZ"), Pauli("IX"), Pauli("XI")])
            observable = Pauli("ZZ")
            ansatz = EfficientSU2(2, reps=1)
            initial_parameters = np.zeros(ansatz.num_parameters)

            time = 1
            optimizer = L_BFGS_B()

            # setup the algorithm
            pvqd = PVQD(
                fidelity,
                ansatz,
                estimator,
                initial_parameters,
                num_timesteps=100,
                optimizer=optimizer,
            )

            # specify the evolution problem
            problem = EvolutionProblem(
                hamiltonian, time, aux_operators=[hamiltonian, observable]
            )

            # and evolve!
            result = pvqd.evolve(problem)

    References:

        [1] Stefano Barison, Filippo Vicentini, and Giuseppe Carleo (2021), An efficient
            quantum algorithm for the time evolution of parameterized circuits,
            `Quantum 5, 512 <https://quantum-journal.org/papers/q-2021-07-28-512/>`_.
    """

    def __init__(
        self,
        fidelity: BaseStateFidelity,
        ansatz: QuantumCircuit,
        initial_parameters: np.ndarray,
        estimator: BaseEstimator | None = None,
        optimizer: Optimizer | Minimizer | None = None,
        num_timesteps: int | None = None,
        evolution: EvolutionSynthesis | None = None,
        use_parameter_shift: bool = True,
        initial_guess: np.ndarray | None = None,
    ) -> None:
        """
        Args:
            fidelity: A fidelity primitive used by the algorithm.
            ansatz: A parameterized circuit preparing the variational ansatz to model the
                time evolved quantum state.
            initial_parameters: The initial parameters for the ansatz. Together with the ansatz,
                these define the initial state of the time evolution.
            estimator: An estimator primitive used for calculating expected values of auxiliary
                operators (if provided via the problem).
            optimizer: The classical optimizers used to minimize the overlap between
                Trotterization and ansatz. Can be either a :class:`.Optimizer` or a callable
                using the :class:`.Minimizer` protocol. This argument is optional since it is
                not required for :meth:`get_loss`, but it has to be set before :meth:`evolve`
                is called.
            num_timesteps: The number of time steps. If ``None`` it will be set such that the
                timestep is close to 0.01.
            evolution: The evolution synthesis to use for the construction of the Trotter step.
                Defaults to first-order Lie-Trotter decomposition, see also
                :mod:`~qiskit.synthesis.evolution` for different options.
            use_parameter_shift: If True, use the parameter shift rule to compute gradients.
                If False, the optimizer will not be passed a gradient callable. In that case,
                Qiskit optimizers will use a finite difference rule to approximate the gradients.
            initial_guess: The initial guess for the first VQE optimization. Afterwards the
                previous iteration result is used as initial guess. If None, this is set to
                a random vector with elements in the interval :math:`[-0.01, 0.01]`.
        """
        super().__init__()
        if evolution is None:
            evolution = LieTrotter()

        self.ansatz = ansatz
        self.initial_parameters = initial_parameters
        self.num_timesteps = num_timesteps
        self.optimizer = optimizer
        self.initial_guess = initial_guess
        self.estimator = estimator
        self.fidelity_primitive = fidelity
        self.evolution = evolution
        self.use_parameter_shift = use_parameter_shift

    def step(
        self,
        hamiltonian: BaseOperator | PauliSumOp,
        ansatz: QuantumCircuit,
        theta: np.ndarray,
        dt: float,
        initial_guess: np.ndarray,
    ) -> tuple[np.ndarray, float]:
        """Perform a single time step.

        Args:
            hamiltonian: The Hamiltonian under which to evolve.
            ansatz: The parameterized quantum circuit which attempts to approximate the
                time-evolved state.
            theta: The current parameters.
            dt: The time step.
            initial_guess: The initial guess for the classical optimization of the
                fidelity between the next variational state and the Trotter-evolved last state.
                If None, this is set to a random vector with elements in the interval
                :math:`[-0.01, 0.01]`.

        Returns:
            A tuple consisting of the next parameters and the fidelity of the optimization.
        """
        self._validate_setup()

        loss, gradient = self.get_loss(hamiltonian, ansatz, dt, theta)

        if initial_guess is None:
            initial_guess = np.random.random(self.initial_parameters.size) * 0.01

        if isinstance(self.optimizer, Optimizer):
            optimizer_result = self.optimizer.minimize(loss, initial_guess, gradient)
        else:
            optimizer_result = self.optimizer(loss, initial_guess, gradient)

        # clip the fidelity to [0, 1]
        fidelity = np.clip(1 - optimizer_result.fun, 0, 1)

        return theta + optimizer_result.x, fidelity

    def get_loss(
        self,
        hamiltonian: BaseOperator | PauliSumOp,
        ansatz: QuantumCircuit,
        dt: float,
        current_parameters: np.ndarray,
    ) -> tuple[Callable[[np.ndarray], float], Callable[[np.ndarray], np.ndarray]] | None:

        """Get a function to evaluate the infidelity between Trotter step and ansatz.

        Args:
            hamiltonian: The Hamiltonian under which to evolve.
            ansatz: The parameterized quantum circuit which attempts to approximate the
                time-evolved state.
            dt: The time step.
            current_parameters: The current parameters.

        Returns:
            A callable to evaluate the infidelity and, if gradients are supported and required,
                a second callable to evaluate the gradient of the infidelity.
        """
        self._validate_setup(skip={"optimizer"})

        # use Trotterization to evolve the current state
        trotterized = ansatz.bind_parameters(current_parameters)

        evolution_gate = PauliEvolutionGate(hamiltonian, time=dt, synthesis=self.evolution)

        trotterized.append(evolution_gate, ansatz.qubits)

        # define the overlap of the Trotterized state and the ansatz
        x = ParameterVector("w", ansatz.num_parameters)
        shifted = ansatz.assign_parameters(current_parameters + x)

        def evaluate_loss(displacement: np.ndarray | list[np.ndarray]) -> float | list[float]:
            """Evaluate the overlap of the ansatz with the Trotterized evolution.

            Args:
                displacement: The parameters for the ansatz.

            Returns:
                The fidelity of the ansatz with parameters ``theta`` and the Trotterized evolution.

            Raises:
                AlgorithmError: If a primitive job fails.
            """
            if isinstance(displacement, list):
                displacement = np.asarray(displacement)
                value_dict = {x_i: displacement[:, i].tolist() for i, x_i in enumerate(x)}
            else:
                value_dict = dict(zip(x, displacement))

            param_dicts = self._transpose_param_dicts(value_dict)
            num_of_param_sets = len(param_dicts)
            states1 = [trotterized] * num_of_param_sets
            states2 = [shifted] * num_of_param_sets
            param_dicts2 = [list(param_dict.values()) for param_dict in param_dicts]
            # the first state does not have free parameters so values_1 will be None by default
            try:
                job = self.fidelity_primitive.run(states1, states2, values_2=param_dicts2)
                fidelities = job.result().fidelities
            except Exception as exc:
                raise AlgorithmError("The primitive job failed!") from exc

            if len(fidelities) == 1:
                fidelities = fidelities[0]

            # in principle, we could add different loss functions here, but we're currently
            # not aware of a use-case for a different one than in the paper
            return 1 - np.abs(fidelities) ** 2

        if _is_gradient_supported(ansatz) and self.use_parameter_shift:

            def evaluate_gradient(displacement: np.ndarray) -> np.ndarray:
                """Evaluate the gradient with the parameter-shift rule.

                This is hardcoded here since the gradient framework does not support computing
                gradients for overlaps.

                Args:
                    displacement: The parameters for the ansatz.

                Returns:
                    The gradient.
                """
                # construct lists where each element is shifted by plus (or minus) pi/2
                dim = displacement.size
                plus_shifts = (displacement + np.pi / 2 * np.identity(dim)).tolist()
                minus_shifts = (displacement - np.pi / 2 * np.identity(dim)).tolist()

                evaluated = evaluate_loss(plus_shifts + minus_shifts)

                gradient = (evaluated[:dim] - evaluated[dim:]) / 2

                return gradient

        else:
            evaluate_gradient = None

        return evaluate_loss, evaluate_gradient

    def _transpose_param_dicts(self, params: dict) -> list[dict[Parameter, float]]:
        p_0 = list(params.values())[0]
        if isinstance(p_0, (list, np.ndarray)):
            num_parameterizations = len(p_0)
            param_bindings = [
                {param: value_list[i] for param, value_list in params.items()}  # type: ignore
                for i in range(num_parameterizations)
            ]
        else:
            param_bindings = [params]

        return param_bindings

    def evolve(self, evolution_problem: TimeEvolutionProblem) -> TimeEvolutionResult:
        r"""Perform real time evolution :math:`\exp(-i t H)|\Psi\rangle`.

        Evolves an initial state :math:`|\Psi\rangle` for a time :math:`t`
        under a Hamiltonian  :math:`H`, as provided in the ``evolution_problem``.

        Args:
            evolution_problem: The evolution problem containing the hamiltonian, total evolution
                time and observables to evaluate.

        Returns:
            A result object containing the evolution information and evaluated observables.

        Raises:
            ValueError: If ``aux_operators`` provided in the time evolution problem but no estimator
                provided to the algorithm.
            NotImplementedError: If the evolution problem contains an initial state.
        """
        self._validate_setup()

        time = evolution_problem.time
        observables = evolution_problem.aux_operators
        hamiltonian = evolution_problem.hamiltonian

        # determine the number of timesteps and set the timestep
        num_timesteps = (
            int(np.ceil(time / 0.01)) if self.num_timesteps is None else self.num_timesteps
        )
        timestep = time / num_timesteps

        if evolution_problem.initial_state is not None:
            raise NotImplementedError(
                "Setting an initial state for the evolution is not yet supported for PVQD."
            )

        # get the function to evaluate the observables for a given set of ansatz parameters
        if observables is not None:
            if self.estimator is None:
                raise ValueError(
                    "The evolution problem contained aux_operators but no estimator was provided. "
                )
            evaluate_observables = _get_observable_evaluator(
                self.ansatz, observables, self.estimator
            )
            observable_values = [evaluate_observables(self.initial_parameters)]

        fidelities = [1]
        parameters = [self.initial_parameters]
        times = np.linspace(0, time, num_timesteps + 1).tolist()  # +1 to include initial time 0

        initial_guess = self.initial_guess

        for _ in range(num_timesteps):
            # perform VQE to find the next parameters
            next_parameters, fidelity = self.step(
                hamiltonian, self.ansatz, parameters[-1], timestep, initial_guess
            )

            # set initial guess to last parameter update
            initial_guess = next_parameters - parameters[-1]

            parameters.append(next_parameters)
            fidelities.append(fidelity)
            if observables is not None:
                observable_values.append(evaluate_observables(next_parameters))

        evolved_state = self.ansatz.bind_parameters(parameters[-1])

        result = PVQDResult(
            evolved_state=evolved_state,
            times=times,
            parameters=parameters,
            fidelities=fidelities,
            estimated_error=1 - np.prod(fidelities),
        )
        if observables is not None:
            result.observables = observable_values
            result.aux_ops_evaluated = observable_values[-1]

        return result

    def _validate_setup(self, skip=None):
        """Validate the current setup and raise an error if something misses to run."""

        if skip is None:
            skip = {}

        required_attributes = {"optimizer"}.difference(skip)

        for attr in required_attributes:
            if getattr(self, attr, None) is None:
                raise ValueError(f"The {attr} cannot be None.")

        if self.num_timesteps is not None and self.num_timesteps <= 0:
            raise ValueError(
                f"The number of timesteps must be positive but is {self.num_timesteps}."
            )

        if self.ansatz.num_parameters == 0:
            raise QiskitError(
                "The ansatz cannot have 0 parameters, otherwise it cannot be trained."
            )

        if len(self.initial_parameters) != self.ansatz.num_parameters:
            raise QiskitError(
                f"Mismatching number of parameters in the ansatz ({self.ansatz.num_parameters}) "
                f"and the initial parameters ({len(self.initial_parameters)})."
            )
